#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for quantum states.
"""

from argparse import ArgumentTypeError
from typing import List, Any
import numpy
from numpy import conj, random, trace, sqrt, reshape, transpose, real, square, abs
from qcompute_qnet import EPSILON
from qcompute_qnet.quantum.gate import Gate

__all__ = [
    "QuantumState",
    "PureState",
    "MixedState",
    "Zero",
    "One",
    "Plus",
    "Minus"
]


class QuantumState:
    r"""Class for creating a quantum state.

    Attributes:
        state (numpy.ndarray): matrix representation of the quantum state
        systems (list): system labels of the quantum state
    """

    def __init__(self, state=None, systems=None):
        r"""Constructor for QuantumState class.

        Args:
            state (numpy.ndarray): initial quantum state
            systems (list): system labels of the quantum state
        """
        self.state = numpy.array([[1]]) if state is None else state
        self.systems = [] if systems is None else systems
        if len(self.systems) != self.size:
            raise ArgumentTypeError("The size of the quantum state and its systems do not match.")

    @property
    def length(self) -> int:
        r"""The length of the matrix of the quantum state.

        Returns:
            int: length of the matrix
        """
        return max(self.state.shape)

    @property
    def size(self) -> int:
        r"""The size of the quantum state.

        Returns:
            int: size of the quantum state
        """
        return int(numpy.log2(self.length))

    def init(self, state: numpy.ndarray) -> None:
        r"""Quantum state initialization.

        Args:
            state (numpy.ndarray): initial quantum state
        """
        self.state = state

    def reset(self) -> None:
        r"""Reset the quantum state.
        """
        self.state = numpy.array([[1]])
        self.systems = []


class PureState(QuantumState):
    r"""Class for pure quantum states.

    Note:
        The matrix representation of a pure state is a column vector.
    """

    def __init__(self, state=None, systems=None):
        r"""Constructor for PureState class.

        Args:
            state (numpy.ndarray): initial quantum state
            systems (list): system labels of the quantum state
        """
        super().__init__(state, systems)
        if len(self.state.shape) != 2:
            raise ArgumentTypeError("The matrix representation of the state is not of two dimensional.")
        elif self.state.shape[1] > 1:
            raise ArgumentTypeError("The matrix shape is not correct!")

    @property
    def ket(self) -> numpy.ndarray:
        r"""Return the ket form of the pure quantum state.

        Returns:
            numpy.ndarray: the ket form of the pure quantum state
        """
        return self.state

    @property
    def bra(self) -> numpy.ndarray:
        r"""Return the bra form of the pure quantum state.

        Returns:
            numpy.ndarray: the bra form of the pure quantum state
        """
        return conj(self.state).T

    @property
    def norm(self):
        r"""Return the norm of the pure quantum state.

        Returns:
            numpy.ndarray: the norm of the pure quantum state
        """
        return sqrt(self.bra @ self.ket)

    @property
    def projector(self) -> numpy.ndarray:
        r"""Transform the state vector to a projector.

        Returns:
            numpy.ndarray: projector of the state vector
        """
        return self.ket @ self.bra

    def evolve(self, which_qubit: list, operator: numpy.ndarray) -> None:
        r"""Evolve a quantum state on a given qubit by the given operator.

        Warning:
            The operator to act should be a two-dimensional square matrix such as a unitary gate matrix.
            The systems to act on should match the dimension of the operator.

        Args:
            which_qubit (list): a list of qubit systems to act on
            operator (numpy.ndarray): operator to act on
        """
        assert isinstance(which_qubit, list), f"Input {which_qubit} should a list."
        assert set(which_qubit).issubset(self.systems), \
            f"Input {which_qubit} does not match the quantum system {self.systems}."
        assert len(operator.shape) == 2, f"Input {operator} should be a two-dimensional matrix."
        assert operator.shape[0] == operator.shape[1], f"Input {operator} is not a square matrix."
        assert 2 ** len(which_qubit) == operator.shape[0], \
            f"Input {which_qubit} does not match the dimension of {operator}."

        self.permute_systems(which_qubit)  # permute the target systems to the front
        shape = [2 ** len(which_qubit), 2 ** (self.size - len(which_qubit))]
        self.state = reshape(operator @ reshape(self.state, shape), [self.length, 1])

    def evolve_by_gates(self, gate_history: list) -> None:
        r"""Evolve the quantum state by a list of gates.

        Note:
            This aims to get the quantum output. So the measurement in the list is skipped.

        Args:
            gate_history (list): a list of quantum gates
        """
        to_matrix = {
            's': Gate.S, 't': Gate.T, 'h': Gate.H, 'x': Gate.X, 'y': Gate.Y, 'z': Gate.Z,
            'u3': Gate.U3, 'rx': Gate.Rx, 'ry': Gate.Ry, 'rz': Gate.Rz,
            'cnot': Gate.CNOT, 'cz': Gate.CZ, 'swap': Gate.SWAP
        }

        for gate in gate_history:
            if gate["name"] in ['s', 't', 'h', 'x', 'y', 'z']:
                self.evolve(gate["which_qubit"], to_matrix[gate["name"]]())
            elif gate["name"] in ['rx', 'ry', 'rz']:
                self.evolve(gate["which_qubit"], to_matrix[gate["name"]](gate["params"]))
            elif gate["name"] in ['u3']:
                self.evolve(gate["which_qubit"], to_matrix[gate["name"]](*gate["params"]))
            elif gate["name"] in ['cnot', 'cz', 'swap']:
                self.evolve(gate["which_qubit"], to_matrix[gate["name"]]())
            elif gate["name"] in ['m']:
                continue

    def sample(self, shots: int) -> dict:
        r"""Sample from the quantum state with a given number of shots.

        This is equivalent to performing quantum measurement in the computational basis.

        Warning:
            Please be aware of the system order and permute the quantum state first if necessary.

        Args:
            shots: the number of samples

        Returns:
            dict: sampling results
        """
        samples = list(random.choice(list(range(self.length)), shots, p=square(abs(self.state)).T[0]))
        counts = {f'{key:0{self.size}b}': samples.count(key) for key in set(samples)}

        return counts

    def measure(self, basis: numpy.ndarray) -> int:
        r"""Measure the quantum state with a given basis.

        Args:
            basis (numpy.ndarray): basis vectors

        Returns:
            int: measurement outcome
        """
        raise NotImplementedError

    def permute_to_front(self, system: Any) -> None:
        r"""Permute a system of the quantum state to the front.

        Args:
            system (any): system to permute
        """
        if system not in self.systems:
            raise ArgumentTypeError(f"The system to permute must be one of the state systems.")

        system_idx = self.systems.index(system)

        if system_idx == 0:
            pass
        elif system_idx == self.size - 1:  # last system
            new_shape = [2 ** (self.size - 1), 2]
            new_axis = [1, 0]
            new_systems = [system] + self.systems[: system_idx]
            self.state = reshape(transpose(reshape(self.state, new_shape), new_axis), [self.length, 1])
            self.systems = new_systems
        else:  # middle system
            new_shape = [2 ** system_idx, 2, 2 ** (self.size - system_idx - 1)]
            new_axis = [1, 0, 2]
            new_systems = [system] + self.systems[: system_idx] + self.systems[system_idx + 1:]
            self.state = reshape(transpose(reshape(self.state, new_shape), new_axis), [self.length, 1])
            self.systems = new_systems

    def permute_systems(self, systems: list):
        r"""Permute the system of quantum state to a given system order.

        Args:
            systems (list): target system order
        """
        for system in reversed(systems):
            self.permute_to_front(system)

    def compare_by_vector(self, other: "PureState") -> float:
        r"""Compare two pure quantum states by their state vectors.

        Args:
            other (PureState): the other pure quantum state to compare

        Returns:
            float: norm difference
        """
        if self.size != other.size:
            raise ArgumentTypeError(f"The two states to compare must have the same size.")

        if not (set(self.systems).issubset(other.systems) and set(other.systems).issubset(self.systems)):
            raise ArgumentTypeError(f"The two states to compare have different system labels,\n"
                                    f"{self.systems}\n"
                                    f"{other.systems}.")

        if abs(self.norm - 1) >= EPSILON:
            raise ArgumentTypeError(f"{self} is not normalized.")
        elif abs(other.norm - 1) >= EPSILON:
            raise ArgumentTypeError(f"{other} is not normalized.")
        else:
            self.permute_systems(other.systems)  # match the system order
            self_state_list = list(self.state)
            # Find an index with the largest absolute value
            idx = self_state_list.index(max(self_state_list, key=abs))
            if abs(other.state[idx]) <= EPSILON:
                error = 1
            else:
                # Calculate the relative phase and erase it
                phase = self.state[idx] / other.state[idx]
                self_phase = self.state / phase
                error = numpy.linalg.norm(self_phase - other.state)
        return error

    def compare_by_density(self, other: "PureState") -> float:
        r"""Compare two pure quantum states by their density matrices.

        Args:
            other (PureState): the second quantum state
        """
        if self.size != other.size:
            raise ArgumentTypeError(f"The two states to compare must have the same size.")

        if not (set(self.systems).issubset(other.systems) and set(other.systems).issubset(self.systems)):
            raise ArgumentTypeError(f"The two states to compare have different system labels,\n"
                                    f"{self.systems}\n"
                                    f"{other.systems}.")

        self.permute_systems(other.systems)  # match the system order
        error = numpy.linalg.norm(self.projector - other.projector)

        return error

    @classmethod
    def random_state_vector(cls, qubit_number: int, is_real=False) -> numpy.ndarray:
        r"""Generate a random pure quantum state vector with given number of qubits.

        Args:
            qubit_number (int): number of qubits
            is_real (int, optional): whether to generate a real matrix

        Returns:
            numpy.ndarray: state vector
        """
        if not isinstance(qubit_number, int):
            raise ArgumentTypeError(f"Input {qubit_number} should be an int value.")

        if not isinstance(is_real, bool):
            raise ArgumentTypeError(f"Input {is_real} should be a bool value.")

        if is_real:
            psi = random.randn(2 ** qubit_number, 1)
            inner_prod = conj(psi).T @ psi
        else:
            psi = random.randn(2 ** qubit_number, 1) + 1j * random.randn(2 ** qubit_number, 1)
            inner_prod = real(conj(psi).T @ psi)

        psi = psi / sqrt(inner_prod)  # normalize the vector
        return psi


class MixedState(QuantumState):
    r"""Class for mixed quantum states.
    """

    def __init__(self, state=None, systems=None):
        r"""Constructor for MixedState class.

        Args:
            state (numpy.ndarray): initial quantum state
            systems (list): system labels of the quantum state
        """
        super().__init__(state, systems)

    def evolve(self, kraus_list: List[numpy.ndarray]) -> None:
        r"""Evolution of the quantum state.

        Args:
            kraus_list (List[numpy.ndarray]): kraus operators acting on the quantum state
        """
        raise NotImplementedError

    def measure(self, basis: numpy.ndarray) -> int:
        r"""Measure the quantum state with a given basis.

        Warning:
            We only consider single-qubit state measurement in this version.

        Args:
            basis (numpy.ndarray): measurement basis

        Returns:
            int: measurement outcome
        """
        prob = [trace(item @ conj(item).T @ self.state).real for item in basis]

        if random.random_sample() < prob[0]:
            self.state = basis[0] @ conj(basis[0]).T
            return 0
        else:
            self.state = basis[1] @ conj(basis[1]).T
            return 1


class Zero:
    r"""Class to obtain a zero state.
    """

    SV = numpy.array([[1], [0]], dtype=complex)
    DM = numpy.array([[1, 0], [0, 0]], dtype=complex)


class One:
    r"""Class to obtain a one state.
    """

    SV = numpy.array([[0], [1]], dtype=complex)
    DM = numpy.array([[0, 0], [0, 1]], dtype=complex)


class Plus:
    r"""Class to obtain a plus state.
    """

    SV = numpy.array([[1], [1]], dtype=complex) / sqrt(2.0)
    DM = numpy.array([[1, 1], [1, 1]], dtype=complex) / 2


class Minus:
    r"""Class to obtain a minus state.
    """

    SV = numpy.array([[1], [-1]], dtype=complex) / sqrt(2.0)
    DM = numpy.array([[1, -1], [-1, 1]], dtype=complex) / 2



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for messages utilized in the qpu module.
"""

from qcompute_qnet.messages.message import QuantumMessage


class QuantumMsg(QuantumMessage):
    r"""Class for the quantum messages utilized in the qpu module.

    Attributes:
        index (int): index of the quantum message
    """

    def __init__(self, data: "QuantumState", index: int):
        r"""Constructor for QuantumMsg class.

        Args:
            data (QuantumState): quantum state of the quantum message
            index (int): index of the quantum message
        """
        super().__init__(data)
        self.index = index
        


#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The backend for measurement-based quantum computation.
"""

from argparse import ArgumentTypeError
from typing import List, Dict, Tuple, Any, Union
from matplotlib import pyplot as plt
import networkx
from networkx import Graph, spring_layout, draw_networkx
import numpy
from numpy import reshape, pi, conj, real, random, sqrt
from qcompute_qnet import EPSILON
from qcompute_qnet.quantum.basis import Basis
from qcompute_qnet.quantum.gate import Gate
from qcompute_qnet.quantum.mcalculus import transpile_to_pattern
from qcompute_qnet.quantum.pattern import Pattern
from qcompute_qnet.quantum.state import PureState, Plus
from qcompute_qnet.quantum.utils import kron, print_progress

__all__ = [
    "MBQC",
    "run_circuit"
]


class MBQC:
    r"""Class for the measurement-based quantum computation.

    Attributes:
        vertex (Vertex): vertices in MBQC algorithm
        max_active (int): maximum number of active vertices
    """

    def __init__(self):
        r"""Constructor for MBQC class.
        """
        self.__graph = None
        self.__pattern = None
        self.vertex = None
        self.max_active = 0  # maximum number of active vertices

        self.__vertex_to_state = {}  # mapping of quantum states and the vertices on the graph
        self.__bg_state = PureState()  # initialize a trivial state
        self.__outcome = {}  # store the measurement outcomes

        self.__history = [self.__bg_state]  # background state history
        self.__status = self.__history[-1] if self.__history != [] else None

        self.__draw = False  # drawing the change of vertex classification
        self.__pause_time = None  # drawing pause time
        self.__track = False  # tracking the computational progress
        self.__pos = None  # vertex positions

    class Vertex:
        r"""Class for vertices in MBQC.

        Each vertex corresponds to a qubit.
        We classify vertices in MBQC to three categories and manipulate them dynamically.
        This helps to run MBQC algorithms in a large scale.

        Attributes:
            total (list): all vertices in the MBQC algorithm
            pending (list): pending vertices to activate
            active (list): active vertices in the current measurement step
            measured (list): measured vertices
        """

        def __init__(self, total=None, pending=None, active=None, measured=None):
            r"""Constructor for Vertex class.

            Args:
                total (list): all vertices in the MBQC algorithm
                pending (list): pending vertices to activate
                active (list): active vertices in the current measurement step
                measured (list): measured vertices
            """
            self.total = [] if total is None else total
            self.pending = [] if pending is None else pending
            self.active = [] if active is None else active
            self.measured = [] if measured is None else measured

    def set_graph(self, graph: List[List]) -> None:
        r"""Set the underlying graph of MBQC algorithm.

        Args:
            graph (List[List]): the underlying graph of MBQC algorithm

        Examples:
            The graph is given by a list as follows.

            >>> mbqc = MBQC()
            >>> V = ['1', '2', '3', '4', '5']  # a list of vertices
            >>> E = [('1', '2'), ('2', '3'), ('3', '4'), ('4', '5')]  # a list of edges
            >>> G = [V, E]
            >>> mbqc.set_graph(G)
        """
        if not isinstance(graph, List):
            raise ArgumentTypeError(f"Input {graph} should be a list.")

        vertices, edges = graph
        if not isinstance(vertices, List):
            raise ArgumentTypeError(f"The first element of {graph} should be a list of vertices.")
        if not isinstance(edges, List):
            raise ArgumentTypeError(f"The second element of {graph} should be a list of edges.")

        vertices_of_edges = set([vertex for edge in edges for vertex in list(edge)])
        if not vertices_of_edges.issubset(vertices):
            raise ArgumentTypeError(f"Invalid graph: ({graph})! The edges must link two vertices on the graph.")

        self.__graph = Graph()
        self.__graph.add_nodes_from(vertices)
        self.__graph.add_edges_from(edges)

        self.vertex = self.Vertex(total=vertices, pending=vertices)
        # Initialize each vertex on the graph with a plus state by default
        self.__vertex_to_state = {vertex: PureState(Plus.SV, [vertex]) for vertex in vertices}

    def get_graph(self) -> networkx.Graph:
        r"""Get the underlying graph of MBQC.

        Returns:
            networkx.Graph: the underlying graph of MBQC
        """
        return self.__graph

    def set_pattern(self, pattern: "Pattern") -> None:
        r"""Set a given measurement pattern.

        Args:
            pattern (Pattern): measurement pattern
        """
        if not isinstance(pattern, Pattern):
            raise ArgumentTypeError(f"Input {pattern} should be a 'Pattern'.")

        self.__pattern = pattern
        cmds = self.__pattern.commands[:]

        # Check if the pattern is of a standard EMC form
        cmd_map = {'E': 1, 'M': 2, 'X': 3, 'Z': 4, 'S': 5}
        cmd_num_wild = [cmd_map[cmd.name] for cmd in cmds]
        cmd_num_standard = cmd_num_wild[:]
        cmd_num_standard.sort(reverse=False)

        if not cmd_num_wild == cmd_num_standard:
            raise ArgumentTypeError(f"Input {pattern} is not a standard EMC pattern.")

        # Set graph according to entanglement commands
        edges = [tuple(cmd.which_qubits) for cmd in cmds if cmd.name == 'E']
        vertices = list(set([vertex for edge in edges for vertex in list(edge)]))
        graph = [vertices, edges]
        self.set_graph(graph)

    def get_pattern(self) -> "Pattern":
        r"""Get the measurement pattern.

        Returns:
            Pattern: a measurement pattern
        """
        return self.__pattern

    def __update_history(self) -> None:
        r"""Update the history and the status of the computation.
        """
        self.__history.append(self.__bg_state)
        self.__status = self.__history[-1]

    def get_history(self) -> list:
        r"""Get the computational history of MBQC.

        Returns:
            list: a list of computational history
        """
        return self.__history

    def set_input_state(self, state=None) -> None:
        r"""Set a given input quantum state.

        Warning:
            Different from the circuit model, MBQC uses plus states by default.

        Args:
            state (PureState): input state to set
        """
        if self.__graph is None:
            raise ArgumentTypeError(f"Please set a 'graph' or 'pattern' before calling 'set_input_state'.")

        if state is not None and not isinstance(state, PureState):
            raise ArgumentTypeError(f"Input {state} should be a 'PureState'.")

        vertices = list(self.__graph.nodes)

        if state is None:
            state_vector = numpy.array([[1]])
            systems = []
        else:
            state_vector = state.state
            # If a pattern is set, map the input state systems to the pattern's input
            if self.__pattern is not None:
                for label in state.systems:
                    if not isinstance(label, int):
                        raise ArgumentTypeError(f"System label {label} should be an int value.")
                    if label < 0:
                        raise ArgumentTypeError(f"System label {label} should be an non-negative integer.")

                systems = [label for label in self.__pattern.input_ if int(label[0]) in state.systems]
            else:
                systems = state.systems

        if not set(systems).issubset(vertices):
            raise ArgumentTypeError(f"Input systems {systems} must be a subset of all vertices on the graph.")

        self.__bg_state = PureState(state_vector, systems)
        self.__update_history()
        self.vertex = self.Vertex(total=vertices,
                                  pending=list(set(vertices).difference(systems)),
                                  active=systems,
                                  measured=[])
        self.max_active = len(self.vertex.active)

    def replace_state_on_vertex(self, vertex: Any, state: "PureState") -> None:
        r"""Replace the state on a given vertex.

        Warning:
            This method should be called after ``set_graph`` or ``set_pattern``.

        Args:
            vertex (Any): vertex to replace
            state (PureState): state to replace
        """
        if not isinstance(state, PureState):
            raise ArgumentTypeError(f"Input {state} should be a 'PureState'.")
        if vertex not in self.vertex.total:
            raise ArgumentTypeError(f"Invalid vertex: ({vertex})! This vertex is not on the graph.")

        self.__vertex_to_state[vertex] = state

    def __apply_cz(self, which_qubits_list: List[Tuple[int, int]]) -> None:
        r"""Apply Controlled-Z gate.

        Warning:
            The qubits to manipulate must be active vertices.

        Args:
            which_qubits_list (list): qubits to manipulate
        """
        for which_qubits in which_qubits_list:
            if not set(which_qubits).issubset(self.vertex.active):
                raise ArgumentTypeError(f"Invalid qubits: ({which_qubits})!\n"
                                        f"The qubits in 'which_qubits_list' must be activated first.")
            qubit1 = which_qubits[0]
            qubit2 = which_qubits[1]
            if qubit1 == qubit2:
                raise ArgumentTypeError(f"Invalid qubits: ({which_qubits})!\n"
                                        f"Control qubit must not be the same as target qubit.")

            # Find the control and target qubits and permute them to the front
            self.__bg_state.permute_to_front(qubit1)
            self.__bg_state.permute_to_front(qubit2)

            new_state = self.__bg_state
            new_state_len = new_state.length
            qua_length = int(new_state_len / 4)
            # Reshape the state, apply CZ and reshape it back
            new_state.state = reshape(Gate.CZ() @ reshape(new_state.state, [4, qua_length]), [new_state_len, 1])

            # Update the order of active vertices and the background state
            self.vertex.active = new_state.systems
            self.__bg_state = PureState(new_state.state, new_state.systems)

    def __apply_pauli_gate(self, gate: str, which_qubit: Any) -> None:
        r"""Apply Pauli gate.

        Args:
            gate (str): name of the Pauli gate
            which_qubit (Any): qubit to manipulate
        """
        self.__bg_state.permute_to_front(which_qubit)
        new_state = self.__bg_state
        new_state_len = new_state.length
        half_length = int(new_state_len / 2)

        if gate == 'X':
            gate_mat = Gate.X()
        elif gate == 'Z':
            gate_mat = Gate.Z()
        else:
            raise ArgumentTypeError(f"Input {gate} should be string 'X' or 'Z'.")

        # Reshape the state, apply X and reshape it back
        new_state.state = reshape(gate_mat @ reshape(new_state.state, [2, half_length]), [new_state_len, 1])
        # Update the order of active vertices and the background state
        self.vertex.active = new_state.systems
        self.__bg_state = PureState(new_state.state, new_state.systems)

    def __create_graph_state(self, which_qubit: Any) -> None:
        r"""Create a graph state based on the current qubit to measure.

        Args:
            which_qubit (any): qubit to measure
        """
        which_qubit_neighbors = set(self.__graph.neighbors(which_qubit))
        neighbors_not_measured = which_qubit_neighbors.difference(set(self.vertex.measured))
        # Create a list of systems and apply cz gates
        cz_list = [(which_qubit, qubit) for qubit in neighbors_not_measured]

        # Obtain the qubits to be activated
        append_qubits = {which_qubit}.union(neighbors_not_measured).difference(set(self.vertex.active))
        # Update active and pending lists
        self.vertex.active += list(append_qubits)
        self.vertex.pending = list(set(self.vertex.pending).difference(self.vertex.active))

        # Compute the new background state vector
        new_bg_state_vector = kron([self.__bg_state.state] +
                                   [self.__vertex_to_state[vertex].state for vertex in append_qubits])

        # Update the background state and apply cz
        self.__bg_state = PureState(new_bg_state_vector, self.vertex.active)
        self.__apply_cz(cz_list)
        self.__draw_process('active', which_qubit)

    def measure(self, which_qubit: Any, basis: numpy.ndarray) -> None:
        r"""Measure a given qubit.

        Args:
            which_qubit (Any): qubit to measure
            basis (numpy.ndarray): measurement basis
        """
        self.__draw_process('measuring', which_qubit)
        self.__create_graph_state(which_qubit)

        if which_qubit not in self.vertex.active:
            raise ArgumentTypeError(f"Invalid qubit: ({which_qubit})! The qubit must be activated before measurement.")

        self.__bg_state.permute_to_front(which_qubit)
        new_bg_state = self.__bg_state
        self.vertex.active = new_bg_state.systems
        half_length = int(new_bg_state.length / 2)

        prob = [0, 0]
        state_unnorm = [0, 0]

        # Calculate the probability and post-measurement states
        for result in [0, 1]:
            basis_dag = conj(basis[result]).T
            # Reshape the state, multiply the basis and reshape it back
            state_unnorm[result] = reshape(basis_dag @ reshape(new_bg_state.state, [2, half_length]), [half_length, 1])
            probability = conj(state_unnorm[result]).T @ state_unnorm[result]
            prob[result] = real(probability) if probability.dtype.name == 'COMPLEX128' else probability

        # Randomly choose a result and its corresponding post-measurement state
        prob_zero = real(prob[0].item())
        prob_one = real(prob[1].item())
        if prob_zero < EPSILON:
            result = 1
            post_state_vector = state_unnorm[1]
        elif prob_one < EPSILON:
            result = 0
            post_state_vector = state_unnorm[0]
        else:
            result = random.choice(2, 1, p=[prob_zero, prob_one]).item()
            # Normalize the state after measurement
            post_state_vector = state_unnorm[result] / sqrt(prob[result])

        # Write the measurement result into the dictionary
        self.__outcome.update({which_qubit: int(result)})
        # Update measured list and active list
        self.vertex.measured.append(which_qubit)
        self.max_active = max(len(self.vertex.active), self.max_active)
        self.vertex.active.remove(which_qubit)

        # Update the background state and the history list
        self.__bg_state = PureState(post_state_vector, self.vertex.active)
        self.__update_history()

        self.__draw_process('measured', which_qubit)

    def sum_outcomes(self, which_qubits: list, add_number=None) -> int:
        r"""Sum the measurement outcome of given qubits.

        Args:
            which_qubits (list): qubits to manipulate
            add_number (int): extra number to add to the summation

        Returns:
            int: summation result
        """
        if add_number is None:
            add_number = 0
        else:
            if not isinstance(add_number, int):
                raise ArgumentTypeError(f"Input {add_number} should be an int value.")

        return sum([self.__outcome[label] for label in which_qubits], add_number)

    def correct_byproduct(self, gate: str, which_qubit: Any, power: int) -> None:
        r"""Correct the byproduct operators.

        Args:
            gate (str): correction type to make
            which_qubit (Any): qubit to correct
            power (int): power of the correction operator
        """
        if gate not in ['X', 'Z']:
            raise ArgumentTypeError(f"Input {gate} should be a string 'X' or 'Z'.")
        if not isinstance(power, int):
            raise ArgumentTypeError(f"Input {power} should be an int value.")

        if power % 2 == 1:
            self.__apply_pauli_gate(gate, which_qubit)

        self.__update_history()

    def __run_cmd(self, cmd: Union[Pattern.CommandM, Pattern.CommandX, Pattern.CommandZ]) -> None:
        r"""Run the given command.

        Args:
            cmd (Union[Pattern.CommandM, Pattern.CommandX, Pattern.CommandZ]): command to run
        """
        if cmd.name not in ['M', 'X', 'Z']:
            raise ArgumentTypeError(f"Invalid command ({cmd}) with the name: ({cmd.name})!\n"
                                    f"Only 'M', 'X' and 'Z' are supported as the command name.")

        if cmd.name == 'M':  # execute measurement commands
            signal_s = self.sum_outcomes(cmd.domain_s)
            signal_t = self.sum_outcomes(cmd.domain_t)

            # The adaptive angle is (-1)^{signal_s} * angle + {signal_t} * pi
            adaptive_angle = (-1) ** signal_s * cmd.angle + signal_t * pi
            self.measure(cmd.which_qubit, Basis.Plane(cmd.plane, adaptive_angle))

        else:  # execute X and Z byproduct correction commands
            power = self.sum_outcomes(cmd.domain)
            self.correct_byproduct(cmd.name, cmd.which_qubit, power)

    def __run_cmd_lst(self, cmd_lst: list, bar_start: int, bar_end: int) -> None:
        r"""Run a list of commands

        Args:
            cmd_lst (list): a list of commands to run
            bar_start (int): start point of the progress bar
            bar_end (int): end point of the progress bar
        """
        for i in range(len(cmd_lst)):
            self.__run_cmd(cmd_lst[i])
            print_progress((bar_start + i + 1) / bar_end, "Pattern Running Progress", self.__track)

    def __flip_outcomes(self, cmd_s_lst: list, bar_start: int, bar_end: int) -> None:
        r"""Flip the measurement outcome from the signal shifting commands.

        Args:
            cmd_s_lst (list): a list of commands to run
            bar_start (int): start point of the progress bar
            bar_end (int): end point of the progress bar
        """
        if not isinstance(cmd_s_lst, List):
            raise ArgumentTypeError(f"Input {cmd_s_lst} should be a list.")

        flip = {}
        for i in range(len(cmd_s_lst)):
            cmd_s = cmd_s_lst[i]
            # Execute signal shifting commands
            if cmd_s.which_qubit not in self.vertex.measured:
                raise ArgumentTypeError(f"Invalid vertex index: ({cmd_s.which_qubit})!\n"
                                        f"This qubit is not measured.")

            power = self.sum_outcomes(cmd_s.domain)
            flip[cmd_s.which_qubit] = (self.__outcome[cmd_s.which_qubit] + power % 2) % 2
            print_progress((bar_start + i + 1) / bar_end, "Pattern Running Progress", self.__track)

        # Update outcome dictionary
        self.__outcome = {i: flip[i] if i in flip.keys() else self.__outcome[i] for i in self.__outcome.keys()}

    def kron_unmeasured_qubits(self) -> None:
        r"""Kronecker product to all unmeasured qubits.

        Warning:
            This method is called when user runs MBQC from a measurement pattern.
        """
        self.__draw = False  # turn off the plot switch
        # As the create_graph_state function would change the measured qubits list, we need to record it
        measured_qubits = self.vertex.measured[:]

        for qubit in list(self.__graph.nodes):
            if qubit not in self.vertex.measured:
                self.__create_graph_state(qubit)
                self.vertex.measured.append(qubit)
                self.max_active = max(len(self.vertex.active), self.max_active)
                self.__bg_state = PureState(self.__bg_state.state, self.vertex.active)

        self.vertex.measured = measured_qubits  # restore the measured qubits

    def run_pattern(self) -> None:
        r"""Run the measurement pattern.

        Warning:
            This method is called after ``set_pattern``.
        """
        if self.__pattern is None:
            raise ArgumentTypeError(f"Invalid pattern: ({self.__pattern})!\n"
                                    f"Please set 'pattern' before calling 'run_pattern'.")

        # Execute measurement commands and correction commands
        cmd_m_lst = [cmd for cmd in self.__pattern.commands if cmd.name == 'M']
        cmd_c_lst = [cmd for cmd in self.__pattern.commands if cmd.name in ['X', 'Z']]
        cmd_s_lst = [cmd for cmd in self.__pattern.commands if cmd.name == 'S']
        bar_end = len(cmd_m_lst + cmd_c_lst + cmd_s_lst)

        self.__run_cmd_lst(cmd_m_lst, 0, bar_end)
        # Activate unmeasured qubits before byproduct corrections
        self.kron_unmeasured_qubits()

        self.__run_cmd_lst(cmd_c_lst, len(cmd_m_lst), bar_end)

        # Flip vertices outcomes according to "commandS"
        self.__flip_outcomes(cmd_s_lst, len(cmd_m_lst + cmd_c_lst), bar_end)

        # The output state's label is messy (e.g. [(2, 0), (0, 1), (1, 3)...]),
        # so we permute the systems in order
        q_output = self.__pattern.output_[1]
        self.__status.permute_systems(q_output)
        self.__bg_state = self.__status

        self.__update_history()

    @staticmethod
    def __map_qubit_to_row(out_lst: list) -> Dict:
        r"""Map the output qubits to row index.

        Args:
            out_lst (list): a list of output qubits

        Returns:
            dict: the relation between output qubits and their row index
        """
        return {qubit[0]: qubit for qubit in out_lst}

    def get_classical_output(self) -> Union[str, dict]:
        r"""Get the measurement outcome of MBQC.

        Note:
            If user sets a measurement pattern, then this returns the measurement outcome of the output qubits.
            If user sets a graph of MBQC, then this returns measurement outcomes of all qubits.

        Returns:
            Union[str, dict]: return the measurement outcomes
        """
        # If the input is a pattern, return the string result equivalent to the circuit model
        if self.__pattern is not None:
            width = len(self.__pattern.input_)
            c_output = self.__pattern.output_[0]
            q_output = self.__pattern.output_[1]
            # Obtain the relationship between row number and the output qubit index
            output_lst = c_output + q_output
            row2qubit = self.__map_qubit_to_row(output_lst)

            # Mark the classical outputs with their measurement outcomes
            # Mark a quantum register with a place holder '_'
            # bit_lst = [str(self.__outcome[row2qubit[i]]) if row2qubit[i] in c_output else '_' for i in range(width)]
            bit_lst = [str(self.__outcome[row2qubit[i]]) if row2qubit[i] in c_output else '' for i in range(width)]
            bit_str = ''.join(bit_lst)
            return bit_str

        # If the input is a graph, return the outcome dictionary
        else:
            return self.__outcome

    def get_quantum_output(self) -> "PureState":
        r"""Get quantum output.

        Returns:
            PureState: quantum state after the MBQC algorithm
        """
        return self.__status

    def __set_position(self, pos: Union[dict, bool]) -> None:
        r"""Set the position of graph plotting.

        Args:
            pos (Union[dict, bool]): position to set
        """
        if isinstance(pos, dict):
            self.__pos = pos
        elif isinstance(pos, bool):
            if pos:
                self.__pos = {}
                for vertex in list(self.__graph.nodes):
                    self.__pos[vertex] = [vertex[1], - vertex[0]]
            else:
                self.__pos = spring_layout(self.__graph)  # use 'spring_layout' method
        else:
            raise ArgumentTypeError(f"Invalid position ({pos}) with the type: ({type(pos)})!\n"
                                    f"Only `Bool` and `Dict` are supported as the type of position.")

    def __draw_process(self, which_process: str, which_qubit: Any) -> None:
        r"""Draw the computational process of MBQC.

        Args:
            which_process (str): which process to plot, can be "measuring", "active" or "measured"
            which_qubit (Any): current vertex to focus on
        """
        if self.__draw:
            if which_process not in ['measuring', 'active', 'measured']:
                raise ArgumentTypeError(f"Invalid process name: ({which_process})!\n"
                                        f"Only `measuring`, 'active' and `measured` are supported as the process name.")

            # Find where the 'which_qubit' is
            if which_qubit in self.vertex.pending:
                pending = self.vertex.pending[:]
                pending.remove(which_qubit)
                vertex_sets = [pending, self.vertex.active, [which_qubit], self.vertex.measured]
            elif which_qubit in self.vertex.active:
                active = self.vertex.active[:]
                active.remove(which_qubit)
                vertex_sets = [self.vertex.pending, active, [which_qubit], self.vertex.measured]
            elif which_qubit in self.vertex.measured:
                vertex_sets = [self.vertex.pending, self.vertex.active, [], self.vertex.measured]
            else:
                raise ArgumentTypeError(f"Invalid vertex: ({which_qubit})! This vertex is not on the graph.")

            # Characterize ancilla vertices
            ancilla_qubits = []
            if self.__pattern is not None:
                for vertex in list(self.__graph.nodes):
                    row = vertex[0]
                    col = vertex[1]
                    # Ancilla vertices do not have integer rows and cols
                    if abs(col - int(col)) >= EPSILON or abs(row - int(row)) >= EPSILON:
                        ancilla_qubits.append(vertex)

            plt.cla()
            plt.title("MBQC Running Process", fontsize=15)
            plt.xlabel("Measuring (RED)  Active (GREEN)  Pending (BLUE)  Measured (GRAY)", fontsize=12)
            plt.grid()
            # mngr = plt.get_current_fig_manager()
            # mngr.window.setGeometry(500, 100, 800, 600)
            colors = ['tab:blue', 'tab:green', 'tab:red', 'tab:gray']
            for j in range(4):
                for vertex in vertex_sets[j]:
                    options = {
                        'nodelist': [vertex],
                        'node_color': colors[j],
                        'node_shape': '8' if vertex in ancilla_qubits else 'o',
                        'with_labels': False,
                        'width': 3,
                    }
                    draw_networkx(self.__graph, self.__pos, **options)
                    ax = plt.gca()
                    ax.margins(0.20)
                    plt.axis('on')
                    ax.set_axisbelow(True)
            plt.pause(self.__pause_time)

    def draw_process(self, draw=True, pos=False, pause_time=0.5) -> None:
        r"""Draw the computational process of MBQC.

        Args:
            draw (bool, optional): whether to draw the process
            pos (Union[bool, Dict], optional): position of the graph
            pause_time (float, optional): refresh time of the plot
        """
        if self.__graph is None:
            raise ArgumentTypeError(f"Please set 'graph' or 'pattern' before calling 'draw_process'.")
        if not isinstance(draw, bool):
            raise ArgumentTypeError(f"Input {draw} should be a bool value.")
        if not isinstance(pos, bool):
            if not isinstance(pos, Dict):
                raise ArgumentTypeError(f"Invalid position ({pos}) with the type: ({type(pos)})!\n"
                                        f"Only `Bool` and `Dict` are supported as the type of position.")
        if pause_time <= 0:
            raise ArgumentTypeError(f"Invalid drawing pause time: ({pause_time})!\n"
                                    f"Drawing pause time must be a positive float value.")

        self.__draw = draw
        self.__pause_time = pause_time

        if self.__draw:
            plt.figure()
            plt.ion()
            self.__set_position(pos)

    def track_progress(self, track=True) -> None:
        r""" Track the progress of MBQC running.

        Args:
            track (bool, optional): whether to track the progress
        """
        if not isinstance(track, bool):
            raise ArgumentTypeError(f"Input {track} should be a bool value.")

        self.__track = track


def run_circuit(circuit: "Circuit", shots=1024, input_state=None, optimize=True) -> dict:
    r"""Run a quantum circuit by its equivalent MBQC model.

    Note:
        This method transpiles a quantum circuit to its equivalent MBQC model first
        and then runs the MBQC pattern to get equivalent sampling result.
        It can work well for large-scale quantum shallow circuits with the option ``optimize=True``.

    Warnings:
        We should check if the circuit has sequential registers first.
        If not, we need to perform remapping before running the circuit.

    Args:
        circuit (Circuit): quantum circuit to run
        shots (int, optional): number of sampling
        input_state (PureState, optional): input quantum state
        optimize (bool): whether to optimize the measurement order

    Returns:
        dict: classical results
    """
    # Check if the circuit has sequential registers
    if circuit.width != max(circuit.occupied_indices) + 1:
        remap_circuit = circuit.copy()
        remap_circuit.remap_indices()
        circuit = remap_circuit

    # Add a layer of H gates to start from zero states
    if input_state is None:
        h_layer = [{"name": 'h', "which_qubit": [i], "params": None} for i in range(circuit.width)]
        new_circuit = circuit.copy()
        new_circuit._history = h_layer + circuit.gate_history
        circuit = new_circuit

    pattern = transpile_to_pattern(circuit, shift_signal=True, optimize=optimize, track=False)

    samples = []

    for shot in range(shots):
        mbqc = MBQC()
        mbqc.set_pattern(pattern)
        mbqc.set_input_state(input_state)
        mbqc.run_pattern()
        c_output = mbqc.get_classical_output()
        samples.append(c_output)

    sample_dict = {}
    for string in list(set(samples)):
        sample_dict[string] = samples.count(string)

    return sample_dict



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for the QKD resource management protocol.
"""

from qcompute_qnet.protocols.resource import RMP, RequestQueue
from qcompute_qnet.models.qkd.message import QKDMessage

__all__ = [
    "QKDRMP"
]


class QKDRMP(RMP):
    r"""Class for the resource management protocol for quantum key distribution.

    Attributes:
        request_queue (RequestQueue): the request queue of the QKDRMP protocol
        confirm_records (dict): records for the received CONFIRM message from repeater nodes
    """

    def __init__(self, name: str, node: "Node", max_volume: int = 10) -> None:
        r"""Constructor for QKDRMP class.

        Args:
            name (str): name of the QKDRMP protocol
            node (Node): the node that holds the request queue
            max_volume (int): maximum number of requests the request queue can accommodate
        """
        from qcompute_qnet.models.qkd.node import RMPRepeaterNode, RMPEndNode

        super().__init__(name)
        self.node = node
        self.request_queue = RequestQueue(self.node, max_volume) if isinstance(self.node, RMPRepeaterNode) else None
        self.confirm_records = {} if isinstance(self.node, RMPEndNode) else None

    def receive_upper(self, upper_protocol: type, **kwargs) -> None:
        r"""Receive a message from an upper protocol.

        Args:
            upper_protocol (type): upper protocol that sends the message
            **kwargs (Any): received keyword arguments
        """
        from qcompute_qnet.models.qkd.routing import QKDRoutingL2
        msg = kwargs['msg']

        if msg.data['type'] == QKDMessage.Type.PATH:
            request_msg = QKDMessage(src=self.node,
                                     dst=msg.dst,
                                     protocol=QKDRoutingL2,
                                     data={'type': QKDMessage.Type.REQUEST,
                                           'req_id': msg.data['req_id'],
                                           'path': msg.data['path'],
                                           'key_num': msg.data['key_num'],
                                           'key_length': msg.data['key_length']})
            self.send_lower(QKDRoutingL2, msg=request_msg)

        elif msg.data['type'] == QKDMessage.Type.RESERVE:
            for node in msg.data['path'][1: -1]:
                self.confirm_records[node] = None

            accept_msg = QKDMessage(src=msg.src,
                                    dst=msg.dst,
                                    protocol=QKDRoutingL2,
                                    data={'type': QKDMessage.Type.ACCEPT,
                                          'req_id': msg.data['req_id'],
                                          'path': msg.data['path'],
                                          'key_num': msg.data['key_num'],
                                          'key_length': msg.data['key_length']})
            self.send_lower(QKDRoutingL2, msg=accept_msg, option='forward')

        elif msg.data['type'] == QKDMessage.Type.ACKNOWLEDGE:
            # Generate an 'ACKNOWLEDGE' message and send it to QKDRoutingL2
            ack_msg = QKDMessage(src=msg.src,
                                 dst=msg.dst,
                                 protocol=QKDRoutingL2,
                                 data={'type': QKDMessage.Type.ACKNOWLEDGE,
                                       'req_id': msg.data['req_id'],
                                       'path': msg.data['path'],
                                       'key_num': msg.data['key_num'],
                                       'key_length': msg.data['key_length']})
            self.send_lower(QKDRoutingL2, msg=ack_msg)

        elif msg.data['type'] == QKDMessage.Type.DONE:
            # Send 'DONE' to QKDRoutingL2
            done_msg = QKDMessage(src=msg.src,
                                  dst=msg.dst,
                                  protocol=QKDRoutingL2,
                                  data={'type': QKDMessage.Type.DONE,
                                        'req_id': msg.data['req_id'],
                                        'path': msg.data['path'],
                                        'key_num': msg.data['key_num'],
                                        'key_length': msg.data['key_length']})
            self.send_lower(QKDRoutingL2, msg=done_msg)

    def receive_lower(self, lower_protocol: "type", **kwargs) -> None:
        r"""Receive a message from a lower protocol.

        Args:
            lower_protocol (type): lower protocol that sends the message
            **kwargs (Any):  received keyword arguments
        """
        from qcompute_qnet.models.qkd.application import QKDAppL4
        from qcompute_qnet.models.qkd.routing import QKDRoutingL2
        from qcompute_qnet.models.qkd.node import RMPEndNode, RMPRepeaterNode

        msg = kwargs['msg']

        if msg.data['type'] == QKDMessage.Type.PATH:
            if isinstance(self.node, RMPRepeaterNode):
                # Forward the 'PATH' to the next hop
                next_hop = self.node.quantum_routing_table[msg.dst]
                if self.node not in msg.data['path']:
                    msg.data['path'].append(self.node)

                if isinstance(next_hop, RMPRepeaterNode):
                    # If the request can be fulfilled, forward the 'REQUEST' message
                    if self.node.key_pools[next_hop][0].fulfill(msg):
                        request_msg = QKDMessage(src=msg.src,
                                                 dst=msg.dst,
                                                 protocol=QKDRoutingL2,
                                                 data={'type': QKDMessage.Type.REQUEST,
                                                       'req_id': msg.data['req_id'],
                                                       'path': msg.data['path'],
                                                       'key_num': msg.data['key_num'],
                                                       'key_length': msg.data['key_length']})
                        self.send_lower(QKDRoutingL2, msg=request_msg)
                    else:
                        self.node.env.logger.info(
                            f"{self.node.name} rejected the QKD request from {msg.src.name} to {msg.dst.name}")

                        reject_msg = QKDMessage(src=self.node,
                                                dst=msg.src,
                                                protocol=QKDRoutingL2,
                                                data={'type': QKDMessage.Type.REJECT,
                                                      'req_id': msg.data['req_id'],
                                                      'path': msg.data['path'],
                                                      'err_type': 1,
                                                      'key_num': msg.data['key_num'],
                                                      'key_length': msg.data['key_length']})
                        self.send_lower(QKDRoutingL2, msg=reject_msg)

                elif isinstance(next_hop, RMPEndNode):
                    # Send the 'REQUEST' to QKDRoutingL2 to forward directly
                    request_msg = QKDMessage(src=msg.src,
                                             dst=msg.dst,
                                             protocol=QKDRoutingL2,
                                             data={'type': QKDMessage.Type.REQUEST,
                                                   'req_id': msg.data['req_id'],
                                                   'path': msg.data['path'],
                                                   'key_num': msg.data['key_num'],
                                                   'key_length': msg.data['key_length']})
                    self.send_lower(QKDRoutingL2, msg=request_msg)

            elif isinstance(self.node, RMPEndNode):
                # Receiver: Create a QKDAppL4 instance and send the key request
                if msg.dst == self.node:
                    self.node.set_qkd_app()
                    # Generate a 'RESERVE' message and send it to QKDAppL4
                    reserve_msg = QKDMessage(src=self.node,
                                             dst=msg.src,
                                             protocol=QKDAppL4,
                                             data={'type': QKDMessage.Type.RESERVE,
                                                   'req_id': msg.data['req_id'],
                                                   'path': msg.data['path'],
                                                   'key_num': msg.data['key_num'],
                                                   'key_length': msg.data['key_length']})
                    self.send_upper(QKDAppL4, msg=reserve_msg, towards='down')

                else:   # operations for missent request
                    self.node.env.logger.info(
                        f"{self.node.name} received a missent QKD request from {msg.src.name} to {msg.dst.name}")
                    msg.data['path'].append(self.node)
                    reject_msg = QKDMessage(src=self.node,
                                            dst=msg.src,
                                            protocol=QKDRoutingL2,
                                            data={'type': QKDMessage.Type.REJECT,
                                                  'req_id': msg.data['req_id'],
                                                  'path': msg.data['path'],
                                                  'err_type': 1,
                                                  'key_num': msg.data['key_num'],
                                                  'key_length': msg.data['key_length']})
                    self.send_lower(QKDRoutingL2, msg=reject_msg)

        elif msg.data['type'] == QKDMessage.Type.RESERVE:
            if self.node == msg.dst:  # operations for request sender
                # Send the 'RESERVE' to QKDAppL4
                reserve_msg = QKDMessage(src=msg.src,
                                         dst=msg.dst,
                                         protocol=QKDAppL4,
                                         data={'type': QKDMessage.Type.RESERVE,
                                               'req_id': msg.data['req_id'],
                                               'path': msg.data['path'],
                                               'key_num': msg.data['key_num'],
                                               'key_length': msg.data['key_length']})
                self.send_upper(QKDAppL4, msg=reserve_msg, towards='up')

            else:  # operations of repeater nodes
                pattern = 0
                push_flag = None

                index = msg.data['path'].index(self.node)
                upstream_node, downstream_node = msg.data['path'][index - 1], msg.data['path'][index + 1]
                for neighbor_node in [downstream_node, upstream_node]:
                    if isinstance(neighbor_node, RMPEndNode):
                        pattern += 1

                # Pattern 0: RMPRepeaterNodes on both sides
                if pattern == 0:
                    # Check if the corresponding key pools on both sides are idle and check their volumes
                    if self.node.key_pools[upstream_node][1].idle is True \
                            and self.node.key_pools[upstream_node][1].fulfill(msg) is True \
                            and self.node.key_pools[downstream_node][0].idle is True \
                            and self.node.key_pools[downstream_node][0].fulfill(msg) is True:
                        # Set the status of corresponding key pools to be busy
                        self.node.key_pools[upstream_node][1].idle = False
                        self.node.key_pools[downstream_node][0].idle = False
                        # Generate an 'ACCEPT' message
                        accept_msg = QKDMessage(src=msg.src,
                                                dst=msg.dst,
                                                protocol=QKDRoutingL2,
                                                data={'type': QKDMessage.Type.ACCEPT,
                                                      'req_id': msg.data['req_id'],
                                                      'path': msg.data['path'],
                                                      'key_num': msg.data['key_num'],
                                                      'key_length': msg.data['key_length']})
                        # Call the QKDRoutingL2 to fetch the keys and calculate the ciphertext
                        self.send_lower(QKDRoutingL2, msg=accept_msg, option='operate', pattern=pattern)
                        # Reset the status of the corresponding key pools
                        self.node.key_pools[upstream_node][1].idle = True
                        self.node.key_pools[downstream_node][0].idle = True

                    else:
                        # Push the request to the request queue
                        msg.data['pattern'] = pattern
                        push_flag = self.request_queue.push(msg)

                # Pattern 1: RMPEndNode on one side and RMPRepeaterNode on the other side
                elif pattern == 1:
                    if isinstance(upstream_node, RMPEndNode):
                        end_peer, repeater_peer, towards, key_pool_index = upstream_node, downstream_node, 'down', 0
                    else:
                        end_peer, repeater_peer, towards, key_pool_index = downstream_node, upstream_node, 'up', 1

                    # Prepare for key generation with the end node
                    generate_msg = QKDMessage(src=self.node,
                                              dst=end_peer,
                                              protocol=QKDRoutingL2,
                                              data={'type': QKDMessage.Type.GENERATE,
                                                    'req_id': msg.data['req_id'],
                                                    'path': msg.data['path'],
                                                    'key_num': msg.data['key_num'],
                                                    'key_length': msg.data['key_length']})
                    self.send_lower(QKDRoutingL2, msg=generate_msg, towards=towards)

                    # Check if the corresponding key pool is idle and check the volume of corresponding key pools
                    if self.node.key_pools[repeater_peer][key_pool_index].idle is True \
                            and self.node.key_pools[repeater_peer][key_pool_index].fulfill(msg) is True:
                        self.node.key_pools[repeater_peer][key_pool_index].idle = False

                        # Generate an 'ACCEPT' message
                        accept_msg = QKDMessage(src=msg.src,
                                                dst=msg.dst,
                                                protocol=QKDRoutingL2,
                                                data={'type': QKDMessage.Type.ACCEPT,
                                                      'req_id': msg.data['req_id'],
                                                      'path': msg.data['path'],
                                                      'key_num': msg.data['key_num'],
                                                      'key_length': msg.data['key_length']})
                        # Call the QKDRoutingL2 to fetch the keys
                        self.send_lower(QKDRoutingL2, msg=accept_msg, option='operate',
                                        towards=towards, pattern=pattern)
                        # Reset the status of the corresponding key pool
                        self.node.key_pools[repeater_peer][key_pool_index].idle = True

                    else:
                        # Push the request to the request queue for waiting
                        msg.data['pattern'] = pattern
                        push_flag = self.request_queue.push(msg)

                # Pattern 2: RMPEndNode on both sides
                elif pattern == 2:
                    # Prepare for key generation with the upstream node
                    generate_msg = QKDMessage(src=self.node,
                                              dst=upstream_node,
                                              protocol=QKDRoutingL2,
                                              data={'type': QKDMessage.Type.GENERATE,
                                                    'req_id': msg.data['req_id'],
                                                    'path': msg.data['path'],
                                                    'key_num': msg.data['key_num'],
                                                    'key_length': msg.data['key_length']})
                    self.send_lower(QKDRoutingL2, msg=generate_msg, towards='down')

                    # Prepare for key generation with the downstream node
                    generate_msg = QKDMessage(src=self.node,
                                              dst=downstream_node,
                                              protocol=QKDRoutingL2,
                                              data={'type': QKDMessage.Type.GENERATE,
                                                    'req_id': msg.data['req_id'],
                                                    'path': msg.data['path'],
                                                    'key_num': msg.data['key_num'],
                                                    'key_length': msg.data['key_length']})
                    self.send_lower(QKDRoutingL2, msg=generate_msg, towards='up')

                if push_flag in [True, None]:
                    # Generate an 'ACCEPT' message and send it to QKDRoutingL2
                    accept_msg = QKDMessage(src=msg.src,
                                            dst=msg.dst,
                                            protocol=QKDRoutingL2,
                                            data={'type': QKDMessage.Type.ACCEPT,
                                                  'req_id': msg.data['req_id'],
                                                  'path': msg.data['path'],
                                                  'key_num': msg.data['key_num'],
                                                  'key_length': msg.data['key_length']})
                    self.send_lower(QKDRoutingL2, msg=accept_msg, option='forward')

                    # Generate an 'CONFIRM' message and send it to QKDRoutingL2
                    confirm_msg = QKDMessage(src=self.node,
                                             dst=msg.src,
                                             protocol=QKDRoutingL2,
                                             data={'type': QKDMessage.Type.CONFIRM,
                                                   'req_id': msg.data['req_id'],
                                                   'path': msg.data['path'],
                                                   'key_num': msg.data['key_num'],
                                                   'key_length': msg.data['key_length']})
                    self.send_lower(QKDRoutingL2, msg=confirm_msg)

                elif push_flag is False:
                    self.node.env.logger.info(f"The key request from {msg.dst.name} to {msg.src.name} was rejected "
                                              f"for the request queue is full")
                    # Generate a 'REJECT' message and send it to QKDRoutingL2
                    reject_msg = QKDMessage(src=self.node,
                                            dst=msg.dst,
                                            protocol=QKDRoutingL2,
                                            data={'type': QKDMessage.Type.REJECT,
                                                  'req_id': msg.data['req_id'],
                                                  'path': msg.data['path'],
                                                  'err_type': 2,
                                                  'key_num': msg.data['key_num'],
                                                  'key_length': msg.data['key_length']})
                    self.send_lower(QKDRoutingL2, msg=reject_msg)

                    # Generate another 'REJECT' message and ask the downstream nodes to cancel the resource reservation
                    cancel_msg = QKDMessage(src=self.node,
                                            dst=msg.src,
                                            protocol=QKDRoutingL2,
                                            data={'type': QKDMessage.Type.REJECT,
                                                  'req_id': msg.data['req_id'],
                                                  'path': list(reversed(msg.data['path'])),
                                                  'err_type': 3,
                                                  'key_num': msg.data['key_num'],
                                                  'key_length': msg.data['key_length']})
                    self.send_lower(QKDRoutingL2, msg=cancel_msg)

        elif msg.data['type'] == QKDMessage.Type.CONFIRM:
            # Save the 'CONFIRM' message
            self.confirm_records[msg.src] = msg
            self.node.env.logger.info(f"{self.node.name} received 'CONFIRM' of request {msg.data['req_id']}"
                                      f" from repeater {msg.src.name}")

        elif msg.data['type'] == QKDMessage.Type.REJECT:
            err_type = msg.data['err_type']

            if err_type in [1, 2]:
                towards = 'up'

            else:
                towards = 'down'
                if isinstance(self.node, RMPRepeaterNode):
                    for req in self.request_queue.requests:
                        # If the request is still in the request queue, remove it from the queue and record the event
                        if req.data['req_id'] == msg.data['req_id']:
                            self.request_queue.requests.remove(req)

                            self.node.keys_rejected += msg.data['key_num']
                            self.node.reqs_rejected += 1
                        # If the request is already processed, record the event
                        else:
                            self.node.keys_invalid += msg.data['key_num']
                            self.node.reqs_invalid += 1
                            self.node.keys_delivered -= msg.data['key_num']
                            self.node.reqs_delivered -= 1

            if self.node in [msg.src, msg.dst]:
                self.send_upper(QKDAppL4, msg=msg, towards=towards)  # inform QKDAppL4 protocol to discard the request

        elif msg.data['type'] == QKDMessage.Type.POP:
            pop_counter = 0  # pop counter of the request queue
            while pop_counter < self.request_queue.max_volume:
                if self.request_queue.current_volume == 0:  # no request in the queue
                    self.node.env.logger.info(f"{self.node.name}'s request queue was empty, no request popped")
                    break
                else:
                    pop_msg = self.request_queue.pop()
                    pattern = pop_msg.data['pattern']
                    index = pop_msg.data['path'].index(self.node)
                    upstream_node, downstream_node = pop_msg.data['path'][index - 1], pop_msg.data['path'][index + 1]

                    if pattern == 0:
                        # Check if the corresponding key pools on both sides are idle and check their volumes
                        if self.node.key_pools[upstream_node][1].idle is True \
                                and self.node.key_pools[upstream_node][1].fulfill(pop_msg) is True \
                                and self.node.key_pools[downstream_node][0].idle is True \
                                and self.node.key_pools[downstream_node][0].fulfill(pop_msg) is True:
                            # Set the status of corresponding key pools to be busy
                            self.node.key_pools[upstream_node][1].idle = False
                            self.node.key_pools[downstream_node][0].idle = False
                            # Generate an 'ACCEPT' message
                            accept_msg = QKDMessage(src=pop_msg.src,
                                                    dst=pop_msg.dst,
                                                    protocol=QKDRoutingL2,
                                                    data={'type': QKDMessage.Type.ACCEPT,
                                                          'req_id': pop_msg.data['req_id'],
                                                          'path': pop_msg.data['path'],
                                                          'key_num': pop_msg.data['key_num'],
                                                          'key_length': pop_msg.data['key_length']})
                            # Call the QKDRoutingL2 to fetch the keys and calculate the ciphertext
                            self.send_lower(QKDRoutingL2, msg=accept_msg, option='operate', pattern=pattern)
                            # Reset the status of the corresponding key pools
                            self.node.key_pools[upstream_node][1].idle = True
                            self.node.key_pools[downstream_node][0].idle = True
                            break

                        else:
                            self.node.env.logger.info(f"The key pools of {self.node.name} couldn't serve for the"
                                                      f" request now, pushed the request back, pattern = {pattern}")

                            self.request_queue.push(pop_msg)  # push the request back
                            pop_counter += 1

                    elif pattern == 1:
                        if isinstance(upstream_node, RMPRepeaterNode):
                            repeater_peer = upstream_node
                            key_pool_index = 1
                            towards = 'up'
                        else:
                            repeater_peer = downstream_node
                            key_pool_index = 0
                            towards = 'down'

                        # Check if the corresponding key pool is idle and check the volume of corresponding key pools
                        if self.node.key_pools[repeater_peer][key_pool_index].idle is True \
                                and self.node.key_pools[repeater_peer][key_pool_index].fulfill(pop_msg) is True:
                            self.node.key_pools[repeater_peer][key_pool_index].idle = False

                            # Generate a 'ACCEPT' message
                            accept_msg = QKDMessage(src=pop_msg.src,
                                                    dst=pop_msg.dst,
                                                    protocol=QKDRoutingL2,
                                                    data={'type': QKDMessage.Type.ACCEPT,
                                                          'req_id': pop_msg.data['req_id'],
                                                          'path': pop_msg.data['path'],
                                                          'key_num': pop_msg.data['key_num'],
                                                          'key_length': pop_msg.data['key_length']})
                            self.send_lower(QKDRoutingL2, msg=accept_msg, option='operate',
                                            towards=towards, pattern=pattern)

                            # Reset the status of the corresponding key pool
                            self.node.key_pools[repeater_peer][key_pool_index].idle = True

                        else:
                            self.node.env.logger.info(f"The key pool of {self.node.name} couldn't serve for the"
                                                      f" request now, pushed the request back, pattern = {pattern}")

                            self.request_queue.push(pop_msg)  # push the request back
                            pop_counter += 1

        elif msg.data['type'] == QKDMessage.Type.CIPHERTEXT:
            # Send the 'CIPHERTEXT' to QKDAppL4
            self.send_upper(QKDAppL4, msg=msg, towards='down')

        elif msg.data['type'] == QKDMessage.Type.ACKNOWLEDGE:
            # Record the request
            self.node.reqs_delivered += 1
            self.node.keys_delivered += msg.data['key_num']

        elif msg.data['type'] == QKDMessage.Type.DONE:
            # Send the 'DONE' to QKDAppL4
            self.send_upper(QKDAppL4, msg=msg, towards='up')

        elif msg.data['type'] == QKDMessage.Type.READY:
            # Send the 'READY' to QKDAppL4
            self.send_upper(QKDAppL4, **kwargs)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for utility functions in discrete-event simulation.
"""

import math
from typing import List
from io import StringIO

__all__ = [
    "show_heap",
    "is_heap"
]


def show_heap(heap: List["Event"]) -> None:
    r"""Print the heap.

    Args:
        heap (List[Event]): heap to show
    """

    width = 180
    output = StringIO()
    last_row = -1
    for i, n in enumerate(heap):
        if i:
            row = int(math.floor(math.log(i + 1, 2)))
        else:
            row = 0
        if row != last_row:
            output.write('\n')
        columns = 2 ** row
        col_width = int(math.floor((width * 1.0) / columns))
        output.write(f"ev {i}: '{n.handler.owner.name}.{n.handler.method}' (t = {n.time}, p = {n.priority})"
                     .center(col_width, " "))
        last_row = row
    print('-' * width + "\nEvent list heap:" + output.getvalue() + '\n' + '-' * width)


def is_heap(list_: list) -> bool:
    r"""Check if the list is a heap.

    Args:
        list_ (list): list to be checked

    Returns:
        bool: whether the list is a heap
    """
    bool_ = []
    for i in range((len(list_) - 3) // 2 + 1):
        bool1 = (list_[i] < list_[2 * i + 1]) or (list_[i] == list_[2 * i + 1])
        bool2 = (list_[i] < list_[2 * i + 2]) or (list_[i] == list_[2 * i + 2])
        bool_.append(bool1 and bool2)
    if len(list_) % 2:
        return all(bool_)
    else:
        return all(bool_) and (list_[(len(list_) - 1) // 2] < list_[-1] or list_[(len(list_) - 1) // 2] == list_[-1])



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for random number generators.
"""

import math
from enum import Enum
from typing import List, Optional
from qcompute_qnet.core.des import Entity
from qcompute_qnet.quantum.circuit import Circuit

__all__ = [
    "RandomNumberGenerator"
]


class RandomNumberGenerator(Entity):
    r"""Class for creating a random number generator.

    The random numbers are generated by running a quantum circuit.

    Important:
        We can switch the ``backend`` to get random numbers from real quantum devices,
        for example, using the QCompute backend ``CloudIoPCAS``.
    """

    def __init__(self, name, env: Optional["DESEnv"] = None):
        r"""Constructor for RandomNumberGenerator class.

        Args:
            name (str): name of the random number generator
            env (DESEnv, optional): discrete-event simulation environment
        """
        super().__init__(name, env)

    def init(self) -> None:
        r"""Random number generator initialization.
        """
        assert self.owner != self, f"The random number generator {self.name} has no owner!"

    @staticmethod
    def random(length: int, backend: Optional[Enum] = None, prob=None) -> List[int]:
        r"""Generate random bits by running a quantum circuit.

        Warning:
            The random bits are generated by preparing a zero state and then performing a Ry gate followed by
            measurement in computational basis. This is just one particular option for random bits' generation with
            given bias and more serious post-processing steps are omitted in this version.
            Also note that we have to run the circuit shot by shot because most backends
            only support to return the collected probability distribution instead of the primitive measurement data.

        Important:
            QCompute provides a more efficient interface for obtaining random bits from real quantum devices.
            Please refer to its latest API documentation for details.

        Args:
            length (int): the number of random bits
            prob (List[float]): probability distribution of random bits
            backend (Enum, optional): backend of the quantum circuit

        Returns:
            List[int]: generated random numbers
        """
        assert prob is None or isinstance(prob, List), "`prob` should be a list."
        assert sum(prob) == 1, "Invalid `prob` input. The sum of `prob` is not equal to one!"
        prob = [0.5, 0.5] if prob is None else prob
        cir = Circuit()
        theta = 2 * math.acos(math.sqrt(prob[0]))
        cir.ry(0, theta)
        cir.measure()
        results = []
        for _ in range(length):
            result = cir.run(shots=1, backend=backend)
            results.append(int(list(result['counts'].keys())[0]))
        return results



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for utility functions used in quantum key distribution simulation.
"""

from typing import List
import matplotlib.pyplot as plt
import networkx
from qcompute_qnet.topology.network import Network
from qcompute_qnet.models.qkd.node import EndNode, BackboneNode, RMPEndNode, RMPRepeaterNode

__all__ = [
    "summary",
    "print_traffic"
]


def summary(network: "Network") -> None:
    r"""Summarize the delivered requests and keys.
    """
    sum_reqs = 0
    sum_keys = 0

    for node in network.nodes:
        if isinstance(node, RMPEndNode):
            sum_reqs += node.reqs_delivered
            sum_keys += node.keys_delivered

    sum_reqs = sum_reqs // 2
    sum_keys = sum_keys // 2

    print(f"{sum_reqs} requests are processed.")
    print(f"{sum_keys} keys are delivered.")


def print_traffic(network: "Network", num_color: int) -> None:
    r"""Print the traffic of the network. The color of a repeater node is determined by the number of
    keys it delivered.

    Args:
        network (Network): network to print traffic
        num_color (int): number of gradient colors
    """

    # Record the delivered keys of all repeater nodes
    delivered_keys_set = {}
    for node in network.nodes:
        if isinstance(node, RMPRepeaterNode):
            delivered_keys_set[node] = node.keys_delivered

    # Get the gradient color list
    def gen_colors(num_color: int) -> List:
        """Get the gradient color list.

        Args:
            num_color (int): number of the colors to generate

        Returns:
            List: generated color list
        """
        values = [int(i * 250 / num_color) for i in range(num_color)]
        color_list = ["#%02x%02x%02x" % (200, int(g), 40) for g in values]
        color_list.reverse()
        return color_list

    color_list = gen_colors(num_color)

    # Set color for each repeater node
    step = (max(delivered_keys_set.values()) - min(delivered_keys_set.values())) // num_color
    assert step != 0, f"Illegal step value: 0"
    color_dict = {}
    for node in network.nodes:
        if isinstance(node, RMPRepeaterNode):
            idx = delivered_keys_set[node] // step
            if idx >= num_color:
                idx = num_color - 1
            color_dict[node] = color_list[idx]

    # Get the quantum topology
    network.get_quantum_topology()
    topology = network.quantum_topology

    edge_options = {"width": 1, "edge_color": "tab:orange", "connectionstyle": "arc3, rad=0.05"}
    plt.rcParams['figure.figsize'] = (16, 12)  # set figure size
    nodes_pos = {node: node.location for node in network.nodes}

    # End nodes
    end_nodes = [node for node in network.nodes if isinstance(node, EndNode)]
    ends_pos = {node: node.location for node in end_nodes}
    ends_options = {"nodelist": end_nodes, "node_size": 200, "node_color": "black"}

    # Backbone nodes
    backbone_nodes = [node for node in network.nodes if isinstance(node, BackboneNode)]
    backbones_pos = {node: node.location for node in backbone_nodes}
    backbones_options = {"nodelist": backbone_nodes, "node_size": 200,
                         "node_color": "blue", "edgecolors": "black"}
    # Labels
    labels = {node: node.name for node in network.nodes}
    labels_pos = {node: (node.location[0], node.location[1] - 0.01) for node in network.nodes}

    networkx.draw_networkx_nodes(topology, ends_pos, **ends_options)  # draw end nodes

    # Draw repeater nodes
    for node in network.nodes:
        if isinstance(node, RMPRepeaterNode):
            repeaters_options = {"nodelist": [node], "node_size": 200,
                                 "node_color": color_dict[node], "edgecolors": "black"}
            networkx.draw_networkx_nodes(topology, {node: node.location}, **repeaters_options)

    networkx.draw_networkx_nodes(topology, backbones_pos, **backbones_options)  # draw backbone nodes
    networkx.draw_networkx_edges(topology, nodes_pos, **edge_options)
    networkx.draw_networkx_labels(topology, labels_pos, labels, font_size=10, font_color="black")

    plt.show()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for building a quantum network.
"""

from typing import List
import networkx
import matplotlib.pyplot as plt
from qcompute_qnet.core.des import DESEnv, Entity
from qcompute_qnet.quantum.circuit import Circuit
from qcompute_qnet.topology.node import Node
from qcompute_qnet.topology.link import Link
from qcompute_qnet.devices.channel import Channel, ClassicalChannel, QuantumChannel
from qcompute_qnet.devices.channel import DuplexChannel, DuplexClassicalFiberChannel, DuplexQuantumFiberChannel

__all__ = [
    "Network"
]


class Network(Entity):
    r"""Class for building a quantum network.

    Attributes:
        classical_topology (networkx.MultiDiGraph): classical topology of the network
        quantum_topology (networkx.MultiDiGraph): quantum topology of the network
        default_circuit (Circuit): the default global quantum circuit that records the evolution of network status
        circuits (List[Circuit]): a list of circuits that record the evolutions of network status

    Note:
        ``Network`` is the class for constructing a quantum network and saving its topology. In a network, there are
        communication nodes and links connecting these nodes generally. So we should install nodes and links
        to a network before running the simulation.

    Examples:
        Here we introduce the procedures of building a quantum network. Firstly, we need to create a network.

        >>> network = Network(name="Network")

        Secondly, we should create nodes and links in a network and connect the nodes.

        >>> alice = Node(name="Alice")
        >>> bob = Node(name="Bob")
        >>> link_ab = Link(name="Link_Alice_Bob", ends=(alice, bob))

        Finally, we should install the nodes and links to the network.

        >>> network.install([alice, bob, link_ab])
    """

    def __init__(self, name: str, env=None):
        """Constructor for the Network class.

        Args:
            name (str): name of the network
            env (DESEnv): discrete-event simulation environment of the network
        """
        super().__init__(name, env)
        self.classical_topology = networkx.MultiDiGraph()
        self.quantum_topology = networkx.MultiDiGraph()
        self.default_circuit = Circuit()
        self.circuits = []

    def init(self) -> None:
        r"""Network initialization.

        The ``init`` method will be triggered by the environment initialization.
        """
        self.get_topology()
        self.assign_routing_table()

    @property
    def nodes(self) -> List[Node]:
        r"""Filter the nodes from the network components.

        Returns:
            List[Node]: a list of nodes
        """
        return [component for component in self.components if isinstance(component, Node)]

    @property
    def links(self) -> List[Link]:
        r"""Filter the links from the network components.

        Returns:
            List[Link]: a list of links
        """
        return [component for component in self.components if isinstance(component, Link)]

    def load_topology_from(self, filename: str) -> None:
        r"""Load a network topology from the json file.

        Args:
            filename (str): name of the file saving a network topology

        Note:
            - The ``filename`` should be given as an absolute path to avoid ``FileNotFoundError``.
            - The ``load_topology_from`` method will create duplex fiber channels for a link which connects two nodes
              in this version.
        """
        import json
        topo = json.load(open(filename))
        nodes = []
        links = []

        from qcompute_qnet.models.qkd.node import EndNode, TrustedRepeaterNode, BackboneNode,\
            RMPRepeaterNode, RMPEndNode
        for node in topo['nodes']:
            if node['type'] == "RMPEndNode":
                nodes.append(RMPEndNode(node['name'],
                                        location=(node['longitude'], node['latitude'])
                                        if node.get('longitude') and node.get('latitude') else None))
            elif node['type'] == "EndNode":
                nodes.append(EndNode(node['name'],
                                     location=(node['longitude'], node['latitude'])
                                     if node.get('longitude') and node.get('latitude') else None))
            elif node['type'] == "RMPRepeaterNode":
                nodes.append(RMPRepeaterNode(node['name'],
                                             location=(node['longitude'], node['latitude'])
                                             if node.get('longitude') and node.get('latitude') else None,
                                             key_pool_volume=node.get('key_pool_volume', 20),
                                             key_length=node.get('key_length', 32)))
            elif node['type'] == "TrustedRepeaterNode":
                nodes.append(TrustedRepeaterNode(node['name'],
                                                 location=(node['longitude'], node['latitude'])
                                                 if node.get('longitude') and node.get('latitude') else None))
            elif node['type'] == "BackboneNode":
                nodes.append(BackboneNode(node['name'],
                                          location=(node['longitude'], node['latitude'])
                                          if node.get('longitude') and node.get('latitude') else None))

            else:
                nodes.append(Node(node['name']))

        for link in topo['links']:
            lk = Link(f"{link['node1']}_{link['node2']}")
            n1 = lk.env.get_node(link['node1'])
            n2 = lk.env.get_node(link['node2'])
            lk.connect(n1, n2)
            # Set a duplex classical channel
            c_distance = link['cchannel']['distance']
            c_loss = link['cchannel']['loss'] if "loss" in link['cchannel'].keys() else 0
            c_delay = link['cchannel']['delay'] if "delay" in link['cchannel'].keys() else None
            ch = DuplexClassicalFiberChannel(f"c_{link['node1']}_{link['node2']}",
                                             distance=c_distance, loss=c_loss, delay=c_delay)
            ch.connect(n1, n2)
            # Set a duplex quantum channel
            q_distance = link['qchannel']['distance']

            if link['qchannel'].get('loss'):
                q_loss = link['qchannel']['loss']
            elif link['qchannel'].get('attenuation'):
                q_loss = link['qchannel']['attenuation'] * q_distance * 1e-3
            else:
                q_loss = None

            q_delay = link['qchannel']['delay'] if "delay" in link['qchannel'].keys() else None
            qh = DuplexQuantumFiberChannel(f"q_{link['node1']}_{link['node2']}",
                                           distance=q_distance, loss=q_loss, delay=q_delay)
            qh.connect(n1, n2)
            lk.install([ch, qh])
            links.append(lk)

        self.install(nodes)
        self.install(links)

    def get_topology(self) -> None:
        r"""Get the classical and quantum topology of the network.
        """
        self.get_classical_topology()
        self.get_quantum_topology()

    def get_classical_topology(self) -> None:
        r"""Get the classical topology of the network.
        """
        self.classical_topology.add_nodes_from(self.nodes)
        self.__add_channels(self.classical_topology, [ClassicalChannel, DuplexClassicalFiberChannel])

    def get_quantum_topology(self) -> None:
        r"""Get the quantum topology of the network.
        """
        self.quantum_topology.add_nodes_from(self.nodes)
        self.__add_channels(self.quantum_topology, [QuantumChannel, DuplexQuantumFiberChannel])

    def __add_channels(self, topology: networkx.MultiDiGraph, channel_types: List) -> None:
        r"""Add channels of given types to the topology.

        Args:
            topology (networkx.MultiDiGraph): topology to add channels
            channel_types (List): specified channel types
        """
        for link in self.links:
            for component in link.components:
                if type(component) in channel_types:
                    if isinstance(component, Channel):
                        topology.add_weighted_edges_from([(component.sender, component.receiver,
                                                           component.get_distance())])
                    elif isinstance(component, DuplexChannel):
                        topology.add_weighted_edges_from([(component.channel1_2.sender,
                                                           component.channel1_2.receiver,
                                                           component.channel1_2.get_distance())])
                        topology.add_weighted_edges_from([(component.channel2_1.sender,
                                                           component.channel2_1.receiver,
                                                           component.channel2_1.get_distance())])

    def assign_routing_table(self) -> None:
        r"""Assign routing tables for trusted repeater nodes in the network.

        Note:
            We use the static routing here and apply the Dijkstra's method to compute the shortest weighted path.
            The classical and quantum routing tables are computed separately.
            Both routing costs depend on the length of the communication channel.
        """
        from qcompute_qnet.models.qkd.node import TrustedRepeaterNode
        for node in self.nodes:
            if isinstance(node, TrustedRepeaterNode):
                for dst in self.nodes:
                    if node != dst:
                        classical_path = networkx.dijkstra_path(self.classical_topology,
                                                                source=node, target=dst, weight='weight')
                        node.classical_routing_table[dst] = classical_path[1]
                        quantum_path = networkx.dijkstra_path(self.quantum_topology,
                                                              source=node, target=dst, weight='weight')
                        node.quantum_routing_table[dst] = quantum_path[1]

    def print_classical_topology(self, geo=False) -> None:
        r"""Print the classical topology of the network.

        Args:
            geo (bool): whether to organize the layout by locations of the nodes
        """
        self.get_classical_topology()

        edge_options = {"width": 1, "style": "dashed", "edge_color": "tab:orange", "connectionstyle": "arc3, rad=0.05"}
        self.__print_topology(self.classical_topology, geo, **edge_options)

    def print_quantum_topology(self, geo=False) -> None:
        r"""Print the quantum topology of the network.

        Args:
            geo (bool): whether to organize the layout by locations of the nodes
        """
        self.get_quantum_topology()

        edge_options = {"width": 1, "edge_color": "tab:orange", "connectionstyle": "arc3, rad=0.05"}
        self.__print_topology(self.quantum_topology, geo, **edge_options)

    def __print_topology(self, topology: networkx.MultiDiGraph, geo: bool, **edge_options) -> None:
        r"""Print the topology of the network.

        Args:
            topology (networkx.MultiDiGraph): topology to print
            geo (bool): whether to organize the layout by locations of the nodes
            **edge_options: keyword arguments for edges
        """
        if geo:  # print the topology with locations
            assert all(node.location is not None for node in self.nodes), f"Not all nodes have a geographical location!"
            plt.rcParams['figure.figsize'] = (16, 12)  # set figure size
            nodes_pos = {node: node.location for node in self.nodes}

            from qcompute_qnet.models.qkd.node import EndNode, TrustedRepeaterNode, BackboneNode
            # End nodes
            end_nodes = [node for node in self.nodes if isinstance(node, EndNode)]
            ends_pos = {node: node.location for node in end_nodes}
            ends_options = {"nodelist": end_nodes, "node_size": 200, "node_color": "black"}
            # Repeater nodes
            repeater_nodes = [node for node in self.nodes if isinstance(node, TrustedRepeaterNode)]
            repeaters_pos = {node: node.location for node in repeater_nodes}
            repeaters_options = {"nodelist": repeater_nodes, "node_size": 200,
                                 "node_color": "blue", "edgecolors": "black"}
            # Backbone nodes
            backbone_nodes = [node for node in self.nodes if isinstance(node, BackboneNode)]
            backbones_pos = {node: node.location for node in backbone_nodes}
            backbones_options = {"nodelist": backbone_nodes, "node_size": 200,
                                 "node_color": "red", "edgecolors": "black"}
            # Labels
            labels = {node: node.name for node in self.nodes}
            labels_pos = {node: (node.location[0], node.location[1] - 0.01) for node in self.nodes}

            networkx.draw_networkx_nodes(topology, ends_pos, **ends_options)  # draw end nodes
            networkx.draw_networkx_nodes(topology, repeaters_pos, **repeaters_options)  # draw repeater nodes
            networkx.draw_networkx_nodes(topology, backbones_pos, **backbones_options)  # draw backbone nodes
            networkx.draw_networkx_edges(topology, nodes_pos, **edge_options)
            networkx.draw_networkx_labels(topology, labels_pos, labels, font_size=10)

        else:
            labels = {node: node.name for node in self.nodes}
            pos = networkx.spring_layout(topology)

            networkx.draw_networkx_nodes(topology, pos, node_size=200, node_color="tab:red")
            networkx.draw_networkx_edges(topology, pos, **edge_options)
            networkx.draw_networkx_labels(topology, pos, labels, font_size=10)

        plt.show()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for discrete-event simulation (DES).
"""

import time
from io import StringIO
from abc import ABC, abstractmethod
from typing import Any, List, Union
from enum import Enum, unique
from heapq import heappush, heappop
import numpy as np
import pandas as pd
from qcompute_qnet.core.log import log_init

__all__ = [
    "Entity",
    "Event",
    "EventHandler",
    "Scheduler",
    "DESEnv"
]


class Entity(ABC):
    r"""Class for creating an entity in the discrete-event simulation.

    Attributes:
        name (str): name of the entity
        env (DESEnv): related discrete-event simulation environment
        owner (Entity): owner of the entity
        components (List[Entity]): components that belong to the entity
        scheduler (Scheduler): an event scheduler
        agenda (List[Event]): events scheduled for the entity
        signed_events (List[Event]): events scheduled by the entity
    """

    def __init__(self, name: str, env=None):
        r"""Constructor for Entity class.

        Args:
            name (str): name of the entity
            env (DESEnv): related discrete-event simulation environment
        """
        self.name = name
        if env is None:
            if DESEnv.get_default_env() is None:
                raise ValueError(f"Default DES environment is None. "
                                 f"Should set a default DES environment first"
                                 f"or assign an explicit DES environment to {self.name}.")
            else:
                self.env = DESEnv.get_default_env()
        else:
            self.env = env
        self.env.entities.append(self)
        self.owner = self
        self.components = []
        self.scheduler = Scheduler(self, self.env)
        self.agenda = []
        self.signed_events = []

    @abstractmethod
    def init(self):
        r"""Abstract method that should be implemented by a specific entity.
        """
        pass

    def attach(self, env: "DESEnv") -> None:
        r"""Attach the entity to a discrete-event simulation environment.

        Args:
            env (DESEnv): discrete-event simulation environment to attach
        """
        assert self.env is None, f"The entity is already attached to '{self.env.name}'."
        self.env = env
        self.env.entities.append(self)

    def detach(self) -> None:
        r"""Detach the entity from its discrete-event simulation environment.

        Note:
            Once an entity is detached from a discrete-event simulation environment, all events scheduled for
            the entity will be cancelled.
        """
        self.env.entities.remove(self)
        # Reset the environment and cancel all scheduled events for the entity
        self.env = None
        for event in self.agenda:
            self.scheduler.cancel(event)

    def install(self, entities: Union["Entity", List["Entity"]]) -> None:
        r"""Install a component or a list of components.

        The components installed will be saved in the ``components`` list of the entity.

        Args:
            entities (Union[Entity, List[Entity]]): component entities to install
        """
        if isinstance(entities, Entity):
            self.__install_entity(entities)
        elif isinstance(entities, List) and all(isinstance(entity, Entity) for entity in entities):
            for entity in entities:
                self.__install_entity(entity)
        else:
            raise ValueError("Should input an entity or a list of entities!")

    def __install_entity(self, other: "Entity") -> None:
        r"""Install a component to the entity.

        Args:
            other (Entity): component entity to install
        """
        assert self.env == other.env, "Cannot install an entity with different environment!"
        assert other not in self.components, f"'{other.name}' has already been installed to '{self.name}'."
        other.owner = self
        self.components.append(other)

    def print_components(self) -> None:
        r"""Print the components of the entity.
        """
        df = pd.DataFrame(columns=["type", "name", "env", "owner"])

        for i, component in enumerate(self.components):
            component_info = pd.DataFrame({"type": component.__class__.__name__,
                                           "name": component.name,
                                           "env": component.env.name,
                                           "owner": component.owner.name},
                                          index=[f"Component {i + 1}"])
            df = pd.concat([df, component_info])

        print(f"\nComponent list of the {self.__class__.__name__}, '{self.name}':\n{df.to_string()}")

    def print_agenda(self) -> None:
        r"""Print the events scheduled for the entity.
        """
        df = Event.events_to_dataframe(self.agenda)
        print(f"\nAgenda of {self.name} (unsorted):\n{df.to_string()}")

    def print_signed_events(self) -> None:
        r"""Print the events scheduled by the entity.
        """
        df = Event.events_to_dataframe(self.signed_events)
        print(f"\nSigned events by {self.name} (unsorted):\n{df.to_string()}")


class Event:
    r"""Class for creating an event in the discrete-event simulation.

    An event is scheduled by a scheduler and will be triggered at a given time.

    Attributes:
        time (int): time of the event to be triggered
        handler (EventHandler): handler for processing the event
        signature (Union[Entity, Protocol]): entity or protocol that creates the event
        priority (int): execution priority of the event, smaller value for higher priority
        status (Status): the status of the event

    Note:
        Once two or more events are scheduled at the same time,
        the order of execution is determined by their priorities.
    """

    def __init__(self, time: int, handler: "EventHandler", signature: Union["Entity", "Protocol"], priority=None):
        r"""Constructor for Event class.

        Args:
            time (int): time of the event to be triggered
            handler (EventHandler): handler for processing the event
            signature (Union[Entity, Protocol]): entity or protocol that creates the event
            priority (int): execution priority of the event, smaller value for higher priority
        """
        self.time = time
        self.handler = handler
        self.signature = signature
        self.priority = float("inf") if priority is None else priority
        self.status = Event.Status.CREATED

    @unique
    class Status(Enum):
        r"""Event status.
        """

        CREATED = "Created"
        SCHEDULED = "Scheduled"
        PROCESSED = "Processed"
        CANCELLED = "Cancelled"

    def cancel(self) -> None:
        r"""Set the event status as cancelled.
        """
        self.status = Event.Status.CANCELLED

    @classmethod
    def is_time_valid(cls, time: Any) -> bool:
        r"""Check if the time input is valid.

        Args:
            time (Any): input time

        Returns:
            bool: a bool value to indicate the validity
        """
        return (isinstance(time, int) and time >= 0) or time == float("inf")

    @classmethod
    def is_priority_valid(cls, priority: Any) -> bool:
        r"""Check if a priority input is valid.

        Args:
            priority (Any): input priority

        Returns:
            bool: a bool value to indicate the validity
        """
        return (isinstance(priority, int) and priority >= 0) or priority == float("inf")

    @classmethod
    def events_to_dataframe(cls, events: List["Event"]) -> "pandas.DataFrame":
        r"""Convert a list of events to dataframe.

        Args:
            events (List[Event]): a list of events to convert

        Returns:
            pandas.DataFrame: dataframe structure of the events list
        """
        df = pd.DataFrame(columns=["time", "priority", "object", "method", "status", "signature"])
        for i, event in enumerate(events):
            event_info = pd.DataFrame({"time": event.time,
                                       "priority": event.priority,
                                       "object": event.handler.owner.name,
                                       "method": event.handler.method,
                                       "status": event.status.value,
                                       "signature": event.signature.name},
                                      index=[f"Event {i + 1}"])
            df = pd.concat([df, event_info])
        return df

    def print(self) -> None:
        r"""Print the event details.
        """
        df = Event.events_to_dataframe([self])
        print(f"\nEvent details:\n{df.to_string()}")

    def __eq__(self, other: "Event") -> bool:
        return (self.time == other.time) and (self.priority == other.priority)

    def __ne__(self, other: "Event") -> bool:
        return (self.time != other.time) or (self.priority != other.priority)

    def __gt__(self, other: "Event") -> bool:
        return (self.time > other.time) or (self.time == other.time and self.priority > other.priority)

    def __lt__(self, other: "Event") -> bool:
        return (self.time < other.time) or (self.time == other.time and self.priority < other.priority)


class EventHandler:
    r"""Class for creating an event handler in the discrete-event simulation.

    Once an event is triggered at a certain time, a pre-specified event handler will be activated to handle the event.

    Attributes:
        owner (Union[Entity, Protocol]): an entity or a protocol to process the event
        method (str): method for a specific process
        params (List[Any]): parameters for processing the method
        **kwargs: keyword arguments for processing the method
    """

    def __init__(self, owner: Union["Entity", "Protocol"], method: str, params=None, **kwargs):
        r"""Constructor for EventHandler class.

        Args:
            owner (Union[Entity, Protocol]): an entity or a protocol to process the event
            method (str): method for specific execution
            params (List[Any]): parameters for processing the method
            **kwargs: keyword arguments for processing the method
        """
        assert params is None or isinstance(params, List), "Should input parameters as a list."

        self.owner = owner
        self.method = method
        self.params = [] if params is None else params
        self.kwargs = kwargs

    def handle(self) -> None:
        r"""Method to execute the event.
        """
        return getattr(self.owner, self.method)(*self.params, **self.kwargs)


class Scheduler:
    r"""Class for creating an event scheduler in the discrete-event simulation.

    Note:
        Only objects of ``Entity`` and ``Protocol`` class contain a scheduler which can schedule events
        to be triggered at a specified time on the timeline.

    Attributes:
        owner (Union[Entity, Protocol]): owner of the scheduler
        env (DESEnv): related discrete-event simulation environment
    """

    def __init__(self, owner: Union["Entity", "Protocol"], env=None):
        r"""Constructor for Scheduler class.

        Args:
            owner (Union[Entity, Protocol]): owner of the scheduler
            env (DESEnv): related discrete-event simulation environment
        """
        self.owner = owner
        self.env = env

    def schedule_at(self, time: int, handler: "EventHandler", priority=None) -> None:
        r"""Schedule an event at a given time.

        Args:
            time (int): time of the event
            handler (EventHandler): handler to process the event
            priority (int): priority of the event
        """
        assert self.env is not None, "Cannot find a DES environment for the scheduler."
        event = Event(time, handler, self.owner, priority)
        self.owner.signed_events.append(event)
        event.handler.owner.agenda.append(event)
        self.env.future_events.push(event)
        self.env.counter_scheduled += 1
        event.status = Event.Status.SCHEDULED

    def schedule_after(self, delay: int, handler: "EventHandler", priority=None) -> None:
        r"""Schedule an event after a time delay.

        Args:
            delay (int): time to delay
            handler (EventHandler): handler to process the event
            priority (int): priority of the event
        """
        self.schedule_at(self.env.now + delay, handler, priority)

    def schedule_now(self, handler: "EventHandler", priority=None) -> None:
        r"""Schedule an event at the current time.

        Args:
            handler (EventHandler): handler to process the event
            priority (int): priority of the event
        """
        self.schedule_at(self.env.now, handler, priority)

    def reschedule(self, event: "Event", time=None, priority=None) -> None:
        r"""Reschedule the time or priority of an event.

        This is implemented by cancelling the previous event and scheduling a new event.

        Args:
            event (Event): event to reschedule
            time (int): new time of the event
            priority (int): new priority of the event
        """
        assert time is not None or priority is not None, "Should either input a new time or a new priority."
        if time is not None:
            assert Event.is_time_valid(time), f"Invalid event time!"
        if priority is not None:
            assert Event.is_priority_valid(priority), f"Invalid event priority!"
        new_time = event.time if time is None else time
        new_priority = event.priority if priority is None else priority
        self.cancel(event)
        self.schedule_at(new_time, event.handler, new_priority)

    @staticmethod
    def cancel(event: "Event") -> None:
        r"""Cancel a scheduled event.

        Args:
            event (Event): event to cancel
        """
        event.cancel()


class DESEnv:
    r"""Class for discrete-event simulation environment.

    Important:
        This is the simulation environment for network simulations.
        We should always instantiate a ``DESEnv`` class to create an environment first before any network setup.

    Examples:
        Create a simulation environment and set it as a default.
        Then any entities created will be automatically attached to this default environment
        unless an explict environment parameter is given.

        >>> env = DESEnv(name="BB84", default=True)

    Attributes:
        name (str): name of the discrete-event simulation environment
        end_time (float): end time of the simulation in picoseconds
        entities (List[Entity]): a list of entities attached to this environment
        future_events (FutureEvents): a list of scheduled events to the timeline
        counter_scheduled (int): counter of scheduled events
        counter_handled (int): counter of handled events
        status (str): status of the simulation environment
        logging (bool): whether to output the simulation log file
        logger (logging.Logger): logger of the simulation environment
        log_config (dict): configuration of the simulation log
    """

    __DEFAULT_ENV = None

    def __init__(self, name: str, default=False):
        r"""Constructor for DESEnv class.

        Args:
            name (str): name of the discrete-event simulation environment
            default (bool): whether to set the current environment as default
        """
        self.name = name
        self.__current_time = 0  # use self.now to access this time
        self.end_time = float("inf")
        self.entities = []
        self.future_events = self.FutureEvents()
        self.counter_scheduled = 0
        self.counter_handled = 0
        self.status = DESEnv.Status.STOPPED
        self.logging = False
        self.logger = None
        self.log_config = {'path': None, 'level': "DEBUG"}
        if default:
            self.set_default_env()

    @unique
    class Status(Enum):
        r"""Simulation status.
        """

        READY = "Ready"
        RUNNING = "Running"
        STOPPED = "Stopped"

    def init(self) -> None:
        r"""Manually initialize the simulation environment.

        Initialize all the entities in the simulation environment.
        """
        for entity in self.entities:
            entity.init()
        self.status = DESEnv.Status.READY  # switch the environment status to 'Ready'

    def set_default_env(self) -> None:
        r"""Set the current environment as default.
        """
        DESEnv.__DEFAULT_ENV = self

    @classmethod
    def get_default_env(cls) -> "DESEnv":
        r"""Get the default discrete-event simulation environment.

        Returns:
            DESEnv: the default discrete-event simulation environment
        """
        return DESEnv.__DEFAULT_ENV

    def get_node(self, name: str) -> "Node":
        r"""Get a node entity by its name.

        Args:
            name (str): name of the entity

        Returns:
            Node: node entity to return
        """
        from qcompute_qnet.topology.node import Node
        for entity in self.entities:
            if isinstance(entity, Node) and entity.name == name:
                return entity

    @property
    def network(self) -> "Network":
        r"""Get the quantum network in this environment.

        Returns:
            Network: network in the discrete-event simulation environment
        """
        from qcompute_qnet.topology.network import Network
        for entity in self.entities:
            if isinstance(entity, Network):
                return entity

    def set_log(self, path=None, level=None) -> None:
        r"""Log settings for the discrete-event simulation.

        Args:
            path (str): name and storing path of the log file
            level (str): logging level of the logger
        """
        if path is not None:
            self.log_config['path'] = path
        if level is not None:
            self.log_config['level'] = level

    @property
    def now(self) -> int:
        r"""Return the current time.

        Returns:
            int: current time of the simulation
        """
        return self.__current_time

    def run(self, end_time=None, logging=False, summary=True) -> None:
        r"""Run the simulation and collect necessary statistics.

        The simulation keeps popping events in the ``future_events`` list and executing them by the assigned handlers.
        The simulation time leaps at discrete timestamps where the events take place.
        The simulation will keep running until the future event list is empty or the end time is reached.

        Args:
            end_time (float): end time of the simulation in picoseconds
            logging (bool): whether to output the simulation log file
            summary (bool): whether to print the simulation report on the terminal
        """
        self.end_time = float("inf") if end_time is None else round(end_time)

        self.logging = logging
        self.logger = log_init(self, self.log_config['path'], self.log_config['level'])

        # Use time.time_ns() instead of time.time() to avoid float type accuracy loss
        sim_start = time.time_ns()
        self.logger.info("-" * 20 + "Simulation starts" + "-" * 20)

        self.status = DESEnv.Status.RUNNING  # switch the environment status to 'Running'

        while not self.future_events.is_empty():
            event = self.future_events.pop()
            # Terminate if the event is scheduled after the end time of the simulation
            # The events scheduled at the end time will still be processed
            if event.time > self.end_time:
                self.__current_time = self.end_time
                self.logger.info(f"Reach end time! No more events will be processed")
                break

            assert self.now <= event.time, f"Inappropriate event time scheduled by {event.handler.owner}"

            # Continue if the current event is cancelled
            if event.status == Event.Status.CANCELLED:
                continue

            self.__current_time = event.time  # time leaps
            event.handler.handle()  # event execution
            event.status = Event.Status.PROCESSED
            self.counter_handled += 1

        self.status = DESEnv.Status.STOPPED  # switch the environment status to 'Stopped'

        sim_end = time.time_ns()
        elapsed_time = sim_end - sim_start

        self.logger.info("-" * 20 + "Simulation ends" + "-" * 20)
        self.logger.info(f"Elapsed wall time: {elapsed_time * 1e-9:.4f}s \t"
                         f" Elapsed virtual time: {self.now * 1e-12:.4f}s")
        self.logger.debug(f"Scheduled events: {self.counter_scheduled} \t Handled events: {self.counter_handled}")
        if summary:
            self.print_summary(elapsed_time)

    def stop(self) -> None:
        r"""Stop the simulation.

        Set the ``end_time`` as ``now`` to end the simulation.
        """
        self.end_time = self.now
        self.logger.info("Simulation is stopped")

    def reset(self) -> None:
        r"""Reset the discrete-event simulation environment.
        """
        raise NotImplementedError("Not available in this version!")

    @staticmethod
    def seed(random_seed: int) -> None:
        r"""Set a random seed.

        Args:
            random_seed (int): random seed for the simulation
        """
        np.random.seed = random_seed

    def print_summary(self, elapsed_time: int) -> None:
        r"""Print the summary information of the simulation.
        """
        string = StringIO()
        string.write("-" * 50 + "\n")
        string.write(f"Simulation status:\t'{self.status.value}'\n\n")
        string.write(f"- Elapsed wall time: {elapsed_time * 1e-9:.4f}s\n")
        string.write(f"- Elapsed virtual time: {self.now * 1e-12:.4f}s\n\n")
        string.write(f"- Scheduled events: {self.counter_scheduled}\n")
        string.write(f"- Handled events: {self.counter_handled}\n")
        if self.logging:
            string.write(f"\nLog records saved in: '{self.logger.handlers[0].baseFilename}'\n")
        string.write("-" * 50)
        print(string.getvalue())

    class FutureEvents:
        r"""Class for creating a future event list in the discrete-event simulation.

        ``FutureEvents`` manages a list of scheduled events during the simulation.
        The events are maintained in a heap structure to fasten the event scheduling.

        Attributes:
            events (List[Event]): a list of the scheduled events
        """

        def __init__(self):
            r"""Constructor for FutureEvents class.
            """
            self.events = []

        def push(self, event: "Event") -> None:
            r"""Push the event into the event heap.

            Args:
                event (Event): new scheduled event
            """
            heappush(self.events, event)

        def pop(self) -> "Event":
            r"""Pop the event with the highest executing priority from the event list.

            Returns:
                Event: event with the highest executing priority
            """
            return heappop(self.events)

        def is_empty(self) -> bool:
            r"""Check if the event list is empty.

            Returns:
                bool: whether the event list is empty
            """
            return True if not self.events else False

        def print(self) -> None:
            r"""Print the future event list.
            """
            df = Event.events_to_dataframe(self.events)
            print(f"\nFuture event list (unsorted):\n{df.to_string()}")



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for the QKD application protocol.
"""

from typing import Tuple

from qcompute_qnet.core.des import EventHandler
from qcompute_qnet.protocols.protocol import Protocol
from qcompute_qnet.models.qkd.resource_management import QKDRMP
from qcompute_qnet.models.qkd.routing import QKDRouting
from qcompute_qnet.models.qkd.message import QKDMessage

__all__ = [
    "QKDApp",
    "QKDAppL4"
]


class QKDApp(Protocol):
    r"""Class for the application protocol in the quantum key distribution.

    Attributes:
        start_time (int): start time of QKDApp protocol
        request_from (Node): node that requests for key generation
        request_to (Node): node that receives a QKD request
        key_num (int): number of requested keys
        key_length (int): length of requested keys
        keys (list): generated keys
        ciphertext_lists (Dict[Node, list]): ciphertext lists received from trusted repeater nodes
    """

    def __init__(self, name: str):
        r"""Constructor for QKDApp class.

        Args:
            name (str): name of the QKDApp protocol
        """
        super().__init__(name)
        self.start_time = 0
        self.request_from = None
        self.request_to = None
        self.key_num = 0
        self.key_length = 0
        self.keys = None
        self.ready_to_remove = False
        self.ciphertext_lists = {}

    def start(self, **kwargs) -> None:
        r"""Start QKDApp protocol.

        Args:
            **kwargs: keyword arguments for initialization
        """
        self.start_time = self.node.env.now
        self.request_to = kwargs['dst']
        self.key_num = kwargs['key_num']
        self.key_length = kwargs['key_length']

        handler = EventHandler(self, "send_request")
        self.scheduler.schedule_now(handler)

    def send_request(self) -> None:
        r"""Send a request to the destination.
        """
        self.node.env.logger.info(f"{self.node.name} started a QKD request with {self.request_to.name}")

        request_msg = QKDMessage(src=self.node,
                                 dst=self.request_to,
                                 protocol=QKDRouting,
                                 data={'type': QKDMessage.Type.REQUEST,
                                       'path': [self.node],
                                       'key_num': self.key_num,
                                       'key_length': self.key_length})
        self.send_lower(QKDRouting, msg=request_msg)

    def receive_lower(self, lower_protocol: type, **kwargs) -> None:
        r"""Receive a message from a lower protocol.

        Args:
            lower_protocol (type): lower protocol that sends the message
            **kwargs: received keyword arguments
        """
        msg = kwargs['msg']

        # Bob: Receive the 'REQUEST' message from Alice
        if msg.data['type'] == QKDMessage.Type.REQUEST:
            # Save sender of the QKD request
            self.request_from = msg.src
            # Append 'self.node' to the path
            path = msg.data['path']
            path.append(self.node)
            self.key_num = msg.data['key_num']
            self.key_length = msg.data['key_length']

            # If the request passed repeaters, set 'ciphertext_lists' and wait for key swapping
            if len(path) > 2:
                for node in path[1: -1]:
                    self.ciphertext_lists[node] = None

            # Bob: Send back an 'ACCEPT' message
            accept_msg = QKDMessage(src=self.node,
                                    dst=self.request_from,
                                    protocol=QKDRouting,
                                    data={'type': QKDMessage.Type.ACCEPT,
                                          'path': path,
                                          'key_num': self.key_num,
                                          'key_length': self.key_length})
            self.send_lower(QKDRouting, msg=accept_msg)

        # Receive the 'READY' message from the lower protocol
        elif msg.data['type'] == QKDMessage.Type.READY:
            # Save the sifted keys
            self.keys = msg.data['sifted_keys']
            finish = msg.data['finish']

            if finish:
                self.finish()

        # Bob: Receive ciphertext list from a repeater node
        elif msg.data['type'] == QKDMessage.Type.CIPHERTEXT:
            self.ciphertext_lists[msg.src] = msg.data['ciphertext_list']  # save incoming ciphertext list

            # Return an 'ACKNOWLEDGE' message to the repeater node for acknowledgement
            ack_msg = QKDMessage(src=self.node,
                                 dst=msg.src,
                                 protocol=QKDRouting,
                                 data={'type': QKDMessage.Type.ACKNOWLEDGE})
            self.send_lower(QKDRouting, msg=ack_msg)

            # If receive all ciphertext lists, begin decryption for swapped keys
            if all(ciphertext_list is not None for ciphertext_list in self.ciphertext_lists.values()):
                swapped_keys = self.keys
                for ciphertext_list in self.ciphertext_lists.values():
                    # Do 'xor' operation for decryption
                    swapped_keys = [key ^ ciphertext for key, ciphertext in zip(swapped_keys, ciphertext_list)]
                self.keys = swapped_keys

                # Send back a 'DONE' message to Alice
                done_msg = QKDMessage(src=self.node,
                                      dst=self.request_from,
                                      protocol=QKDRouting,
                                      data={'type': QKDMessage.Type.DONE})
                self.send_lower(QKDRouting, msg=done_msg)
                self.finish()

        # Alice: Receive the 'DONE' message from Bob
        elif msg.data['type'] == QKDMessage.Type.DONE:
            self.finish()

    def finish(self) -> None:
        r"""Finish the QKDApp protocol.
        """
        # Alice: Finish her protocol
        if self.request_to is not None:
            # Deliver the keys to the node
            self.node.keys[self.request_to] = self.keys
            self.node.env.logger.info(f"{self.node.name} generated end-to-end keys with {self.request_to.name}")
            self.node.env.logger.info(f"{self.node.name}'s keys (with {self.request_to.name} in decimal): {self.keys}")
            print(f"{self.node.name}'s keys (with {self.request_to.name} in decimal): {self.keys}")
            self.key_rate_estimation()
            if self.node in self.request_to.keys.keys():
                self.error_rate_estimation()  # estimate error rate

        # Bob: Finish his protocol
        elif self.request_from is not None:
            self.node.keys[self.request_from] = self.keys
            self.node.env.logger.info(f"{self.node.name} generated end-to-end keys with {self.request_from.name}")
            self.node.env.logger.info(f"{self.node.name}'s keys "
                                      f"(with {self.request_from.name} in decimal): {self.keys}")
            print(f"{self.node.name}'s keys (with {self.request_from.name} in decimal): {self.keys}")

            if self.node in self.request_from.keys.keys():
                self.error_rate_estimation()  # estimate error rate

        # Remove the protocol itself from the protocol stack
        self.owner.stack.remove_node(self)
        self.owner = self

    def key_rate_estimation(self) -> float:
        r"""Estimate key rate of generated keys.

        Returns:
            float: estimated end-to-end key rate
        """
        key_rate = (self.key_num * self.key_length / 1000) / ((self.node.env.now - self.start_time) * 1e-12)
        self.node.env.logger.info(f"End-to-end key rate: {key_rate:.4f} kbit/s")
        print(f"End-to-end key rate: {key_rate:.4f} kbit/s")
        return key_rate

    def error_rate_estimation(self) -> float:
        r"""Estimate error rate of generated keys.

        Returns:
            float: estimated error rate of the end-to-end keys
        """
        # Calculate error rate of all generated keys
        num_errors = 0
        for i in range(self.key_num):
            if self.request_to is not None:
                key_xor = self.keys[i] ^ self.request_to.keys[self.node][i]  # in decimal
            elif self.request_from is not None:
                key_xor = self.keys[i] ^ self.request_from.keys[self.node][i]  # in decimal
            num_errors += bin(key_xor).count('1')  # count the number of errors
        error_rate = num_errors / (self.key_num * self.key_length)
        self.node.env.logger.info(f"End-to-end error rate: {error_rate:.4f}")
        print(f"End-to-end error rate: {error_rate:.4f}")
        return error_rate

    def statistics_estimation(self) -> Tuple[float, float]:
        r"""Estimate key rate and error rate of generated keys.

        Returns:
            Tuple[float, float]: estimated key rate and error rate of the end-to-end keys
        """
        key_rate = self.key_rate_estimation()
        error_rate = self.error_rate_estimation()
        return key_rate, error_rate


class QKDAppL4(QKDApp):
    r"""Class for the application protocol in the resource management architecture for quantum key distribution.

    Attributes:
        req_id (str): identifier of the QKD request
    """

    def __init__(self, name: str):
        r"""Constructor for QKDAppL4 class.

        Args:
            name (str): name of the QKDAppL4 protocol
        """
        super().__init__(name)
        self.req_id = None

    def create_request_id(self) -> str:
        r"""Create a unique identifier for the request using md5 algorithm.

        Returns:
            str: identifier for the request (hexadecimal representation)
        """
        import random
        import hashlib

        salt = random.randint(999, 102400)
        self.req_id = hashlib.md5(str(self.node.env.now + salt).encode("utf-8")).hexdigest()

        return self.req_id

    def send_request(self) -> None:
        r"""Send a key request to the destination node.
        """
        self.node.env.logger.info(f"{self.node.name} started a QKD request with {self.request_to.name}")

        request_id = self.create_request_id()   # create a unique identifier for the request
        request_msg = QKDMessage(src=self.node,
                                 dst=self.request_to,
                                 protocol=QKDRMP,
                                 data={'type': QKDMessage.Type.PATH,
                                       'req_id': request_id,
                                       'path': [self.node],
                                       'key_num': self.key_num,
                                       'key_length': self.key_length})
        self.send_lower(QKDRMP, msg=request_msg)

    def receive_lower(self, lower_protocol: type, **kwargs) -> None:
        r"""Receive a message from the lower protocol.

        Args:
            lower_protocol (type): lower protocol that sends the message
            **kwargs: received keyword arguments
        """
        msg = kwargs['msg']
        towards = kwargs['towards']
        if towards == 'up' and self.request_from is None:
            pass
        elif towards == 'down' and self.request_to is None:
            pass
        else:
            return

        if msg.data['type'] == QKDMessage.Type.RESERVE:
            # Receiver: Save related parameters and prepare for key distribution
            if self.node == msg.src:
                req_id = msg.data.get('req_id')
                if self.req_id is None and req_id is not None:
                    self.req_id = req_id

                self.request_from = msg.dst
                self.key_num = msg.data['key_num']
                self.key_length = msg.data['key_length']

                path = msg.data['path']
                # Initialize the ciphertext list for key swapping
                msg.data['path'].append(self.node)
                if len(path) > 2:
                    for node in path[1: -1]:
                        self.ciphertext_lists[node] = None

                # Send the 'RESERVE' message to the lower QKDRMP protocol
                self.send_lower(QKDRMP, msg=msg)

            # Sender: Record the reception of the 'RESERVE' message
            elif self.node == msg.dst:
                self.node.env.logger.info(f"{self.node.name} received 'RESERVE' from {msg.src.name}")

        elif msg.data['type'] == QKDMessage.Type.REJECT:
            self.discard()  # discard the request and record the event.
            if msg.data['err_type'] == 3:
                self.ciphertext_lists = {}  # reset the ciphertext list

        elif msg.data['type'] == QKDMessage.Type.READY:
            # Save the generated keys
            self.keys = msg.data['sifted_keys']
            finish = msg.data['finish']

            if finish:
                self.node.keys_delivered += len(self.keys)
                self.finish()

        # Save the ciphertext and send an 'ACKNOWLEDGE' message to QKDRMP protocol
        elif msg.data['type'] == QKDMessage.Type.CIPHERTEXT:
            # Save the received ciphertexts
            self.ciphertext_lists[msg.src] = msg.data['ciphertext_list']
            self.node.env.logger.info(f"{self.node.name} saved 'CIPHERTEXT' from {msg.src.name}")
            # Send an 'ACKNOWLEDGE' message to QKDRMP protocol
            ack_msg = QKDMessage(src=self.node,
                                 dst=msg.src,
                                 protocol=QKDRMP,
                                 data={'type': QKDMessage.Type.ACKNOWLEDGE,
                                       'req_id': msg.data['req_id'],
                                       'path': msg.data['path'],
                                       'key_num': msg.data['key_num'],
                                       'key_length': msg.data['key_length']})
            self.send_lower(QKDRMP, msg=ack_msg)

            # Decrypt the ciphertext list if all the ciphertexts are received
            if all(ciphertext_list is not None for ciphertext_list in self.ciphertext_lists.values()):
                swapped_keys = self.keys

                for ciphertext_list in self.ciphertext_lists.values():
                    swapped_keys = [key ^ ciphertext for key, ciphertext in zip(swapped_keys, ciphertext_list)]
                    self.keys = swapped_keys

                # Record the request to the delivered list
                self.node.keys_delivered += len(self.keys)

                # Send a 'DONE' message to QKDRMP
                done_msg = QKDMessage(src=self.node,
                                      dst=self.request_from,
                                      protocol=QKDRMP,
                                      data={'type': QKDMessage.Type.DONE,
                                            'req_id': msg.data['req_id'],
                                            'path': msg.data['path'],
                                            'key_num': msg.data['key_num'],
                                            'key_length': msg.data['key_length']})
                self.send_lower(QKDRMP, msg=done_msg)

                self.finish()

        elif msg.data['type'] == QKDMessage.Type.DONE:
            # Record the delivery and end the request
            self.node.env.logger.info(f"{self.node.name} received 'DONE' from {msg.src.name}")
            self.node.keys_delivered += len(self.keys)
            self.finish()

    def discard(self) -> None:
        r"""Discard the request and record the event.
        """
        self.node.reqs_discarded += 1

        # Reset the status of sender and receiver as idle
        self.node.set_idle()
        peer = self.request_from if self.request_from is not None else self.request_to
        peer.set_idle()
        self.node.env.logger.info(f"{self.node.name} discarded the request to {peer.name}")

        # Remove the QKDAppL4 instance
        self.owner.stack.remove_node(self)
        self.owner = self

        self.node.random_req()  # schedule another request

    def finish(self) -> None:
        r"""Finish the QKDAppL4 protocol.
        """
        self.node.env.logger.info(f"{self.node.name} finished the request {self.req_id}")
        self.node.reqs_delivered += 1

        # Sender
        if self.request_to is not None:
            # Deliver the keys to the node
            self.node.keys[self.request_to] = self.keys
            self.node.env.logger.info(f"{self.node.name} generated end-to-end keys with {self.request_to.name}")
            self.node.env.logger.info(f"{self.node.name}'s keys (with {self.request_to.name} in decimal): {self.keys}")
            _, dual_qkd_app = self.error_rate_estimation()  # estimate error rate

        # Receiver
        elif self.request_from is not None:
            self.node.keys[self.request_from] = self.keys
            self.node.env.logger.info(f"{self.node.name} generated end-to-end keys with {self.request_from.name}")
            self.node.env.logger.info(f"{self.node.name}'s keys "
                                      f"(with {self.request_from.name} in decimal): {self.keys}")

            _, dual_qkd_app = self.error_rate_estimation()  # estimate error rate

        # Remove both QKDAppL4 instances if the request is finished at both ends
        if dual_qkd_app.ready_to_remove:
            self.owner.stack.remove_node(self)
            self.owner = self
            dual_qkd_app.owner.stack.remove_node(dual_qkd_app)
            dual_qkd_app.owner = dual_qkd_app

            self.node.set_idle()
            dual_qkd_app.node.set_idle()

            # Generate another key request
            if self.node.random_request:
                self.node.random_req()

    def error_rate_estimation(self) -> Tuple[float, "QKDAppL4"]:
        r"""Estimate error rate of generated end-to-end keys.

        Returns:
            Tuple: a tuple containing estimated error rate of the end-to-end keys and related QKDAppL4 instance
        """
        # Calculate error rate of all generated keys
        num_errors = 0
        # Find the other end node
        dual_node = self.request_to if self.request_to is not None else self.request_from

        # Find the corresponding QKDAppL4 instance of the other end node
        dual_qkd_app = None
        for proto in dual_node.protocol_stack.protocols:
            if isinstance(proto, QKDAppL4):
                if proto.req_id == self.req_id:
                    dual_qkd_app = proto
                    break

        self.ready_to_remove = True

        if dual_qkd_app.keys is None:
            return 0.0, dual_qkd_app

        for i in range(self.key_num):
            key_xor = self.keys[i] ^ dual_qkd_app.keys[i]
            num_errors += bin(key_xor).count('1')  # count the number of errors

        error_rate = num_errors / (self.key_num * self.key_length)

        self.node.env.logger.info(f"End-to-end error rate between {self.node.name} and {dual_node.name}:"
                                  f" {error_rate:.4f}")

        return error_rate, dual_qkd_app



#!/usr/bin/python3
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Baidu, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
Module for photon detectors.
"""

from typing import List, Tuple
from numpy import random
from qcompute_qnet.core.des import Entity, EventHandler
from qcompute_qnet.devices.beamsplitter import PolarizationBeamSplitter

__all__ = [
    "SinglePhotonDetector",
    "PolarizationDetector"
]


class SinglePhotonDetector(Entity):
    r"""Class for creating a single photon detector.

    Attributes:
        _is_working (bool): whether the single photon detector is working
        efficiency (float): probability that successfully detect a photon when it arrives
        dark_count (float): frequency in hertz that the detector is triggered in the absence of any incident photons
        count_rate (float): maximum photon detection rate in hertz
        resolution (int): how accurately the time in picoseconds the detector can record
        _valid_time (int): next valid time for photon detection
    """

    def __init__(self, name: str, env=None, efficiency=0.4, dark_count=0.0, count_rate=10e6, resolution=1):
        r"""Constructor for SinglePhotonDetector class.

        Args:
            name (str): name of the single photon detector
            env (DESEnv): discrete-event simulation environment
            efficiency (float): probability that successfully detect a photon when it arrives
            dark_count (float): frequency in hertz that the detector is triggered in the absence of any incident photons
            count_rate (float): maximum photon detection rate in hertz
            resolution (int): how accurately the time in picoseconds the detector can record
        """
        super().__init__(name, env)
        self._is_working = False
        self.efficiency = efficiency
        self.dark_count = dark_count
        self.count_rate = count_rate
        self.resolution = resolution
        self._valid_time = 0
        self.__records = []

    def init(self) -> None:
        r"""Detector initialization.
        """
        assert self.owner != self, f"The single photon detector {self.name} has no owner!"

    def turn_on(self) -> None:
        r"""Turn on the single photon detector.
        """
        self._is_working = True
        if self.dark_count > 0:
            delay = round(random.exponential(1 / self.dark_count) * 1e12)
            self.scheduler.schedule_after(delay, EventHandler(self, "_receive_dark_count"))

    def turn_off(self) -> None:
        r"""Turn off the single photon detector.
        """
        self._is_working = False
        self.__records = []

    def is_ready(self) -> bool:
        r"""Check if the detector is ready for detection.

        Returns:
            bool: ready or not
        """
        return self.env.now >= self._valid_time

    def _receive_dark_count(self) -> None:
        r"""Simulate the dark count events.
        """
        pass

    def set(self, **kwargs) -> None:
        r"""Set given parameters.

        Args:
            **kwargs: keyword arguments to set
        """
        for attr in kwargs:
            if attr == "efficiency":
                assert isinstance(kwargs[attr], float), "'efficiency' should be a float value."
                self.__setattr__(attr, kwargs[attr])
            elif attr == "dark_count":
                assert isinstance(kwargs[attr], float), "'dark_count' should be a float value."
                self.__setattr__(attr, kwargs[attr])
            elif attr == "count_rate":
                assert isinstance(kwargs[attr], float), "'count_rate' should be a float value."
                self.__setattr__(attr, kwargs[attr])
            elif attr == "resolution":
                assert isinstance(kwargs[attr], int), f"'resolution' should be an int value."
                self.__setattr__(attr, kwargs[attr])
            else:
                raise TypeError(f"Setting {attr} is not allowed in {self.name}")

    def print_parameters(self) -> None:
        r"""Print parameters of the single photon detector.
        """
        print("-" * 50)
        print(f"Details of Single Photon Detector: {self.name}\n"
              f"efficiency: {self.efficiency}\n"
              f"dark count: {self.dark_count} hertz\n"
              f"count rate: {self.count_rate} hertz\n"
              f"resolution: {self.resolution} picoseconds")

    def receive(self, photon_index: int) -> None:
        r"""Receive an incoming photon and record its index and arrival time.

        Args:
            photon_index (int): index of the incoming photon in a pulse
        """
        if self._is_working:
            if random.random_sample() < self.efficiency and self.is_ready():
                self.__records.append((photon_index, round(self.env.now / self.resolution) * self.resolution))
                self._valid_time = self.env.now + round(1e12 / self.count_rate)

    def pop_records(self) -> List[Tuple]:
        r"""Pop the detection records.

        Returns:
            List[Tuple]: photon indices and time records of photon detection
        """
        records, self.__records = self.__records, []
        return records


class PolarizationDetector(Entity):
    r"""Class for polarization detector.

    Attributes:
        detectors (Tuple): single photon detector components of the polarization detector
        beamsplitter (BeamSplitter): beam splitter component of the polarization detector
    """

    def __init__(self, name: str, env=None):
        r"""Constructor for PolarizationDetector class.

        Args:
            name (str): name of the polarization detector
            env (DESEnv): discrete-event simulation environment
        """
        super().__init__(name, env)
        self.detectors = (SinglePhotonDetector(name + ".SPD0", env),
                          SinglePhotonDetector(name + ".SPD1", env))
        self.beamsplitter = PolarizationBeamSplitter(name + ".PBS", env)
        self.install(list(self.detectors))
        self.install(self.beamsplitter)
        self.beamsplitter.set(receivers=self.detectors)

    def init(self) -> None:
        r"""Detector initialization.
        """
        assert self.owner != self, f"The polarization detector {self.name} has no owner!"

    def set_beamsplitter(self, **kwargs) -> None:
        r"""Set given parameters for the beam splitter.

        Args:
            **kwargs: keyword arguments to set
        """
        self.beamsplitter.set(**kwargs)

    def set_detectors(self, **kwargs) -> None:
        r"""Set given parameters for the two single photon detectors.

        Args:
            **kwargs: keyword arguments to set
        """
        self.detectors[0].set(**kwargs)
        self.detectors[1].set(**kwargs)

    def receive(self, photon: "Photon") -> None:
        r"""Receive an incoming photon.

        Args:
            photon (Photon): photon to receive
        """
        self.beamsplitter.receive(photon)

    def print_parameters(self) -> None:
        r"""Print parameters of the polarization detector.
        """
        self.detectors[0].print_parameters()
        self.detectors[1].print_parameters()
        self.beamsplitter.print_parameters()
